from functools import partial
from mypy_extensions import Arg
from typing import Callable, List, Dict, Union


__all__ = [
    "check_success_base",
    "simple_inventory_based_check",
    "simple_stat_kill_entity_based_check",
    "time_since_death_check",
    "use_any_item_check",
    "use_all_item_check",
]


# takes an initial info dict, a current info dict, and elapsed time-steps, return successful or not
check_success_base = Callable[
    [
        Arg(dict, "ini_info_dict"),
        Arg(dict, "cur_info_dict"),
        Arg(int, "elapsed_timesteps"),
    ],
    bool,
]


def _simple_stat_kill_entity_based_check(
    name: str,
    quantity: int,
    ini_info_dict: dict,
    cur_info_dict: dict,
    elapsed_timesteps: int,
):
    """
    A simple success check based on `info["stat"]["kill_entity"][{name}]`.
    """
    return (
        cur_info_dict["stat"]["kill_entity"][name]
        - ini_info_dict["stat"]["kill_entity"][name]
    ) >= quantity


def simple_stat_kill_entity_based_check(
    name: str, quantity: int, **kwargs
) -> check_success_base:
    return partial(_simple_stat_kill_entity_based_check, name=name, quantity=quantity)


def _simple_inventory_based_check(
    name: str,
    quantity: int,
    ini_info_dict: dict,
    cur_info_dict: dict,
    elapsed_timesteps: int,
):
    """
    A simple success check based on `info["inventory"]`
    """
    return (
        sum(
            [
                inv_item["quantity"]
                for inv_item in cur_info_dict["inventory"]
                if inv_item["name"] == name
            ]
        )
        - sum(
            [
                inv_item["quantity"]
                for inv_item in ini_info_dict["inventory"]
                if inv_item["name"] == name
            ]
        )
    ) >= quantity


def simple_inventory_based_check(
    name: str, quantity: int, **kwargs
) -> check_success_base:
    return partial(_simple_inventory_based_check, name=name, quantity=quantity)


def _time_since_death_check(
    threshold, ini_info_dict: dict, cur_info_dict: dict, elapsed_timesteps: int
):
    """
    Success check based on info["time_since_death"]
    """
    return cur_info_dict["stat"]["time_since_death"] >= threshold


def time_since_death_check(threshold, **kwargs) -> check_success_base:
    return partial(_time_since_death_check, threshold=threshold)


def _use_any_item_check(
    targets: Dict[str, int],
    ini_info_dict: dict,
    cur_info_dict: dict,
    elapsed_timesteps: int,
):
    """
    success check based on increment in info["stat"]["use_item"]["minecraft"][item]
    satisfaction of any item will result in "True" -- the logic "any"
    """
    return any(
        [
            (
                cur_info_dict["stat"]["use_item"]["minecraft"][item]
                - ini_info_dict["stat"]["use_item"]["minecraft"][item]
            )
            >= target
            for item, target in targets.items()
        ]
    )


def use_any_item_check(targets: Dict[str, int]) -> check_success_base:
    return partial(_use_any_item_check, targets=targets)


def _use_all_item_check(
    targets: Dict[str, int],
    ini_info_dict: dict,
    cur_info_dict: dict,
    elapsed_timesteps: int,
):
    """
    success check based on increment in info["stat"]["use_item"]["minecraft"][item]
    satisfaction of all item will result in "True" -- the logic "all"
    """
    return all(
        [
            (
                cur_info_dict["stat"]["use_item"]["minecraft"][item]
                - ini_info_dict["stat"]["use_item"]["minecraft"][item]
            )
            >= target
            for item, target in targets.items()
        ]
    )


def use_all_item_check(targets: Dict[str, int]) -> check_success_base:
    return partial(_use_all_item_check, targets=targets)



from __future__ import annotations
import re
import sys
import importlib_resources
from itertools import product
from omegaconf import OmegaConf

from .meta.base import MetaTaskBase
from ..sim import MineDojoSim, InventoryItem
from ..sim.wrappers import FastResetWrapper, ARNNWrapper
from .meta import (
    HarvestMeta,
    CombatMeta,
    TechTreeMeta,
    Playthrough,
    SurvivalMeta,
    CreativeMeta,
)
import logging

_logger = logging.getLogger(__name__)
_logger.setLevel(logging.INFO)
_stream_handler = logging.StreamHandler(stream=sys.stderr)
_stream_handler.setFormatter(logging.Formatter("[%(levelname)s:%(name)s] %(message)s"))
_logger.addHandler(_stream_handler)


def _resource_file_path(fname) -> str:
    with importlib_resources.path("minedojo.tasks.description_files", fname) as p:
        return str(p)


_MetaTaskName2Class = {
    "Open-Ended": MineDojoSim,
    "Harvest": HarvestMeta,
    "Combat": CombatMeta,
    "TechTree": TechTreeMeta,
    "Playthrough": Playthrough,
    "Survival": SurvivalMeta,
    "Creative": CreativeMeta,
}
MetaTaskName2Class = {k.lower(): v for k, v in _MetaTaskName2Class.items()}


def _meta_task_make(meta_task: str, *args, **kwargs) -> MetaTaskBase | FastResetWrapper:
    """
    Gym-style making tasks from names.

    Usage example:

    .. highlight:: python
    .. code-block:: python

        import minedojo
        env = minedojo.make(task_name, *args, **kwargs)

    Args:
        meta_task: Name of the meta task. Can be one of (and their lower-cased equivalents)
            ``"Open-Ended"``, ``"Harvest"``, ``"Combat"``, ``"TechTree"``,
            ``"Playthrough"``, ``"Survival"``, ``"Creative"``.

        *args: See corresponding task's docstring for more info.

        **kwargs: See corresponding task's docstring for more info.
    """
    meta_task = meta_task.lower()
    assert (
        meta_task in MetaTaskName2Class
    ), f"Invalid meta task name provided {meta_task}"
    if meta_task == "open-ended":
        if "fast_reset" in kwargs:
            fast_reset = kwargs.pop("fast_reset")
            fast_reset_random_teleport_range = kwargs.pop(
                "fast_reset_random_teleport_range", None
            )
            if fast_reset is True:
                return FastResetWrapper(
                    MineDojoSim(*args, **kwargs), fast_reset_random_teleport_range
                )
    return MetaTaskName2Class[meta_task](*args, **kwargs)


_ALL_TASKS_SPECS_UNFILLED = OmegaConf.load(_resource_file_path("tasks_specs.yaml"))
# check no duplicates
assert len(set(_ALL_TASKS_SPECS_UNFILLED.keys())) == len(_ALL_TASKS_SPECS_UNFILLED)

# all possible variables used to fill specs
_ALL_VARS = {
    # Combat
    "combat_biomes": ["forest", "plains", "extreme_hills"],
    "regular_biomes_mob": [
        "cow",
        "pig",
        "sheep",
        "chicken",
    ],
    "regular_biomes_night_mob": [
        "zombie",
        "spider",
        "skeleton",
        "creeper",
        "witch",
        "enderman",
    ],
    "end_mob": ["shulker", "endermite"],
    "nether_mob": [
        "blaze",
        "ghast",
        "wither_skeleton",
        "zombie_pigman",
    ],
    "plains_mob": ["horse", "donkey"],
    "weapon_material": ["wooden", "iron", "diamond"],
    "armor_material": ["leather", "iron", "diamond"],
    # Harvest
    "quantity": [1, 8],
    ## wool and milk
    "cow_biomes": ["plains", "extreme_hills", "forest"],
    "sheep_biomes": ["plains", "extreme_hills", "forest"],
    ## mine
    "ore_type": ["iron_ore", "gold_ore", "diamond", "redstone", "coal", "cobblestone"],
    "pickaxe_material": ["wooden", "stone", "iron", "golden", "diamond"],
    ## most supported items (default only)
    "natural_items": [
        "nether_star",
        "blaze_rod",
        "ghast_tear",
        "nether_wart",
        "netherrack",
        "soul_sand",
        "chorus_flower",
        "chorus_fruit",
        "chorus_plant",
        "elytra",
        "end_stone",
        "ender_pearl",
        "apple",
        "beef",
        "beetroot",
        "beetroot_seeds",
        "bone",
        "brown_mushroom",
        "cactus",
        "carrot",
        "chicken",
        "dirt",
        "egg",
        "feather",
        "fish",
        "grass",
        "leaves",
        "log",
        "monster_egg",
        "mutton",
        "porkchop",
        "potato",
        "prismarine_shard",
        "pumpkin",
        "rabbit",
        "red_mushroom",
        "reeds",
        "sapling",
        "skull",
        "snowball",
        "spawn_egg",
        "sponge",
        "string",
        "totem_of_undying",
        "vine",
        "web",
        "wheat_seeds",
        "wheat",
    ],
    "craft_items": [
        "book",
        "carrot_on_a_stick",
        "clay",
        "crafting_table",
        "dye",
        "end_bricks",
        "end_rod",
        "ender_eye",
        "flint_and_steel",
        "glowstone",
        "gold_nugget",
        "iron_nugget",
        "iron_trapdoor",
        "lever",
        "nether_brick",
        "planks",
        "pumpkin_seeds",
        "red_nether_brick",
        "sandstone",
        "shears",
        "slime_ball",
        "stick",
        "stone_button",
        "stonebrick",
        "sugar",
        "torch",
        "trapped_chest",
        "wooden_button",
        "wool",
        "stone_pressure_plate",
    ],
    "crafting_table_items": [
        "anvil",
        "arrow",
        "banner",
        "beacon",
        "bed",
        "beetroot_soup",
        "boat",
        "bookshelf",
        "bowl",
        "bread",
        "bucket",
        "cake",
        "cauldron",
        "chest",
        "cookie",
        "end_crystal",
        "ender_chest",
        "fence",
        "fence_gate",
        "fire_charge",
        "fishing_rod",
        "flower_pot",
        "furnace",
        "glass_bottle",
        "glass_pane",
        "golden_apple",
        "hopper",
        "iron_bars",
        "ladder",
        "lead",
        "map",
        "minecart",
        "mushroom_stew",
        "painting",
        "paper",
        "pumpkin_pie",
        "rabbit_stew",
        "rail",
        "sea_lantern",
        "shield",
        "sign",
        "speckled_melon",
        "stone_slab",
        "trapdoor",
        "tripwire_hook",
        "wooden_door",
        "writable_book",
    ],
    "furnace_items": [
        "baked_potato",
        "brick",
        "cooked_beef",
        "cooked_chicken",
        "cooked_fish",
        "cooked_mutton",
        "cooked_porkchop",
        "cooked_rabbit",
        "glass",
        "gold_ingot",
        "iron_ingot",
        "quartz",
        "stone",
        "emerald",
        "netherbrick",
    ],
    ## core items
    ### most of the items here need trees in the biomes
    "biome_subset": ["plains", "jungle", "taiga", "forest", "swampland"],
    "natural_core": [
        "apple",
        "beef",
        "bone",
        "chicken",
        "log",
        "reeds",
        "web",
        "wheat",
    ],
    "hand_craft_core": [
        "flint_and_steel",
        "crafting_table",
        "planks",
        "shears",
        "stick",
        "sugar",
        "torch",
    ],
    "crafting_table_core": [
        "arrow",
        "chest",
        "shield",
        "fishing_rod",
        "bucket",
        "furnace",
    ],
    "furnace_core": [
        "cooked_beef",
        "glass",
        "gold_ingot",
        "iron_ingot",
        "brick",
        "stone",
    ],
    # Tech-tree
    "from_barehand_tools": ["wooden", "stone"],
    "from_barehand_tools_armor": ["iron", "golden", "diamond"],
    "from_wood_tools": ["stone"],
    "from_wood_tools_armor": ["iron", "golden", "diamond"],
    "from_stone_tools_armor": [
        "iron",
        "golden",
        "diamond",
    ],
    "from_iron_tools_armor": [
        "golden",
        "diamond",
    ],
    "from_gold_tools_armor": [
        "diamond",
    ],
    "target_tools": [
        "sword",
        "pickaxe",
        "axe",
        "hoe",
        "shovel",
    ],
    "target_armor": [
        "boots",
        "chestplate",
        "helmet",
        "leggings",
    ],
    "target_tools_armor": [
        "sword",
        "pickaxe",
        "axe",
        "hoe",
        "shovel",
        "boots",
        "chestplate",
        "helmet",
        "leggings",
    ],
    "redstone_list": [
        "redstone_block",
        "clock",
        "compass",
        "dispenser",
        "dropper",
        "observer",
        "piston",
        "redstone_lamp",
        "redstone_torch",
        "repeater",
        "detector_rail",
        "comparator",
        "activator_rail",
    ],
}


def product_dict(**kwargs):
    keys = kwargs.keys()
    vals = kwargs.values()
    for instance in product(*vals):
        yield dict(zip(keys, instance))


ALL_TASKS_SPECS = {}
for task_id, task_specs in _ALL_TASKS_SPECS_UNFILLED.items():
    unfilled_vars = re.findall(r"\{(.*?)\}", task_id)

    def _recursive_find_unfilled_vars(x):
        if OmegaConf.is_dict(x):
            return {k: _recursive_find_unfilled_vars(v) for k, v in x.items()}
        elif isinstance(x, str):
            unfilled_vars.extend(re.findall(r"\{(.*?)\}", x))
        return x

    _recursive_find_unfilled_vars(task_specs)

    # deduplicate unfilled vars
    unfilled_vars = list(set(unfilled_vars))
    if len(unfilled_vars) == 0:
        # no unfilled vars, just make the task
        ALL_TASKS_SPECS[task_id] = task_specs
    else:
        unfilled_vars_values = {var: _ALL_VARS[var] for var in unfilled_vars}
        for var_dict in product_dict(**unfilled_vars_values):
            filled_task_id = task_id.format(**var_dict)

            def _recursive_replace_var(x):
                if OmegaConf.is_dict(x):
                    return {k: _recursive_replace_var(v) for k, v in x.items()}
                elif isinstance(x, str):
                    return x.format(**var_dict)
                return x

            task_specs_filled = _recursive_replace_var(task_specs)

            for k, v in task_specs_filled.items():
                if k == "target_quantities":
                    task_specs_filled[k] = int(v)
            task_specs_filled["prompt"] = task_specs_filled["prompt"].replace("_", " ")
            task_specs_filled["prompt"] = task_specs_filled["prompt"].replace(" 1", "")

            ALL_TASKS_SPECS[filled_task_id] = task_specs_filled

# check no duplicates
assert len(set(ALL_TASKS_SPECS.keys())) == len(ALL_TASKS_SPECS)

# load prompts and guidance for programmatic tasks
P_TASKS_PROMPTS_GUIDANCE = OmegaConf.load(
    _resource_file_path("programmatic_tasks.yaml")
)
# check no duplicates
assert len(set(P_TASKS_PROMPTS_GUIDANCE.keys())) == len(P_TASKS_PROMPTS_GUIDANCE)

# load prompts and guidance for creative tasks
C_TASKS_PROMPTS_GUIDANCE = OmegaConf.load(_resource_file_path("creative_tasks.yaml"))
# check no duplicates
assert len(set(C_TASKS_PROMPTS_GUIDANCE.keys())) == len(C_TASKS_PROMPTS_GUIDANCE)

# load prompt and guidance for Playthrough task
PLAYTHROUGH_PROMPT_GUIDANCE = OmegaConf.load(
    _resource_file_path("playthrough_task.yaml")
)
# check only one playthrough task
assert len(PLAYTHROUGH_PROMPT_GUIDANCE.keys()) == 1

ALL_PROGRAMMATIC_TASK_IDS = list(P_TASKS_PROMPTS_GUIDANCE.keys())
ALL_PROGRAMMATIC_TASK_INSTRUCTIONS = {
    task_id: (
        P_TASKS_PROMPTS_GUIDANCE[task_id]["prompt"],
        P_TASKS_PROMPTS_GUIDANCE[task_id]["guidance"],
    )
    for task_id in ALL_PROGRAMMATIC_TASK_IDS
}
ALL_CREATIVE_TASK_IDS = list(C_TASKS_PROMPTS_GUIDANCE.keys())
ALL_CREATIVE_TASK_INSTRUCTIONS = {
    task_id: (
        C_TASKS_PROMPTS_GUIDANCE[task_id]["prompt"],
        C_TASKS_PROMPTS_GUIDANCE[task_id]["guidance"],
    )
    for task_id in ALL_CREATIVE_TASK_IDS
}
PLAYTHROUGH_TASK_ID = list(PLAYTHROUGH_PROMPT_GUIDANCE.keys())[0]
PLAYTHROUGH_TASK_INSTRUCTION = (
    PLAYTHROUGH_PROMPT_GUIDANCE[PLAYTHROUGH_TASK_ID]["prompt"],
    PLAYTHROUGH_PROMPT_GUIDANCE[PLAYTHROUGH_TASK_ID]["guidance"],
)
ALL_TASK_IDS = ALL_PROGRAMMATIC_TASK_IDS + ALL_CREATIVE_TASK_IDS + [PLAYTHROUGH_TASK_ID]
ALL_TASK_INSTRUCTIONS = {
    **ALL_PROGRAMMATIC_TASK_INSTRUCTIONS,
    **ALL_CREATIVE_TASK_INSTRUCTIONS,
    PLAYTHROUGH_TASK_ID: PLAYTHROUGH_TASK_INSTRUCTION,
}

_logger.info(
    f"Loaded {len(P_TASKS_PROMPTS_GUIDANCE)} Programmatic tasks, "
    f"{len(C_TASKS_PROMPTS_GUIDANCE)} Creative tasks, "
    """and 1 special task: "Playthrough". """
    f"Totally {len(P_TASKS_PROMPTS_GUIDANCE) + len(C_TASKS_PROMPTS_GUIDANCE) + 1} tasks loaded."
)


def _parse_inventory_dict(inv_dict: dict[str, dict]) -> list[InventoryItem]:
    return [InventoryItem(slot=k, **v) for k, v in inv_dict.items()]


def _specific_task_make(task_id: str, *args, **kwargs):
    assert task_id in ALL_TASKS_SPECS, f"Invalid task id provided {task_id}"
    task_specs = ALL_TASKS_SPECS[task_id]

    # handle list of inventory items
    if "initial_inventory" in task_specs:
        kwargs["initial_inventory"] = _parse_inventory_dict(
            task_specs["initial_inventory"]
        )
        task_specs.pop("initial_inventory")

    # pop prompt from task specs because it is set from programmatic yaml
    task_specs.pop("prompt")

    # meta task
    meta_task_cls = task_specs.pop("__cls__")
    task_obj = _meta_task_make(meta_task_cls, *args, **task_specs, **kwargs)
    return task_obj


def make(task_id: str, *args, cam_interval: int | float = 15, **kwargs):
    """
    Make a task. task_id can be one of the following:
    1. a task id for Programmatic tasks, e.g., "combat_bat_extreme_hills_barehand"
    2. format "creative:{idx}" for the idx-th Creative task
    3. "playthrough" or "open-ended" for these two special tasks
    4. one of "harvest", "combat", "techtree", and "survival" to creative meta task
    """
    if task_id.startswith("creative:"):
        creative_idx = int(task_id.split(":")[1])
        assert len(C_TASKS_PROMPTS_GUIDANCE) > creative_idx >= 0
        info = C_TASKS_PROMPTS_GUIDANCE[task_id]
        env_obj = _meta_task_make("creative", *args, **kwargs)
        env_obj.specify_prompt(info["prompt"])
        env_obj.specify_guidance(info["guidance"])
        env_obj.collection = info["collection"]
        env_obj.source = info["source"]
    elif task_id in P_TASKS_PROMPTS_GUIDANCE:
        info = P_TASKS_PROMPTS_GUIDANCE[task_id]
        env_obj = _specific_task_make(task_id, *args, **kwargs)
        env_obj.specify_prompt(info["prompt"])
        env_obj.specify_guidance(info["guidance"])
    elif task_id.lower() == PLAYTHROUGH_TASK_ID.lower():
        info = PLAYTHROUGH_PROMPT_GUIDANCE[PLAYTHROUGH_TASK_ID]
        env_obj = _specific_task_make(task_id, *args, **kwargs)
        env_obj.specify_prompt(info["prompt"])
        env_obj.specify_guidance(info["guidance"])
    elif task_id.lower() in [
        "open-ended",
        "harvest",
        "combat",
        "techtree",
        "survival",
    ]:
        env_obj = _meta_task_make(task_id, *args, **kwargs)
    else:
        raise ValueError(f"Invalid task id provided {task_id}")
    return ARNNWrapper(env_obj, cam_interval=cam_interval)



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
from typing import List

import numpy as np

import minedojo.sim.mc_meta.mc as mc
from minedojo.sim import spaces
from minedojo.sim.handlers.translation import (
    KeymapTranslationHandler,
    TranslationHandlerGroup,
)


__all__ = ["ObserveFromFullStats"]


class ObserveFromFullStats(TranslationHandlerGroup):
    """
    Includes the use_item statistics for every item in MC that can be used
    """

    def xml_template(self) -> str:
        return str("""<ObservationFromFullStats/>""")

    def to_string(self) -> str:
        return self.stat_key

    def __init__(self, stat_key=None):
        if stat_key is None:
            self.stat_key = "full_stats"
            super(ObserveFromFullStats, self).__init__(
                handlers=[
                    _FullStatsObservation(statKeys)
                    for statKeys in mc.ALL_STAT_KEYS
                    if len(statKeys) == 2
                ]
            )
        else:
            self.stat_key = stat_key
            super(ObserveFromFullStats, self).__init__(
                handlers=[
                    _FullStatsObservation(statKeys)
                    for statKeys in mc.ALL_STAT_KEYS
                    if statKeys[1] == stat_key
                ]
            )


class _FullStatsObservation(KeymapTranslationHandler):
    def to_hero(self, x) -> int:
        for key in self.hero_keys:
            x = x[key]
        return x

    def __init__(self, key_list: List[str], space=None, default_if_missing=None):
        if space is None:
            if "achievement" == key_list[0]:
                space = spaces.Box(low=0, high=1, shape=(), dtype=int)
            else:
                space = spaces.Box(low=0, high=np.inf, shape=(), dtype=int)
        if default_if_missing is None:
            default_if_missing = np.zeros((), dtype=float)

        super().__init__(
            hero_keys=key_list,
            univ_keys=key_list,
            space=space,
            default_if_missing=default_if_missing,
        )

    def xml_template(self) -> str:
        return str("""<ObservationFromFullStats/>""")



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
import numpy as np

from minedojo.sim import spaces
from minedojo.sim.handlers.translation import (
    KeymapTranslationHandler,
    TranslationHandlerGroup,
)


__all__ = ["CompassObservation"]


class CompassObservation(TranslationHandlerGroup):
    """
    Defines compass observations.

    Args:
        angle (bool, optional): Whether or not to include angle observation. Defaults to True.
        distance (bool, optional): Whether or not ot include distance observation. Defaults to False.

    Example usage:

    .. code-block:: python

        # A compass observation object which gives angle and distance information
        CompassObservation(True, True)
    """

    def to_string(self) -> str:
        return "compass"

    def xml_template(self) -> str:
        return str("""<ObservationFromCompass/>""")

    def __init__(self, angle=True, distance=False):
        """Initializes a compass observation."""

        assert angle or distance, "Must observe either angle or distance"

        handlers = []

        if angle:
            handlers.append(_CompassAngleObservation())
        if distance:
            handlers.append(
                KeymapTranslationHandler(
                    hero_keys=["distanceToCompassTarget"],
                    univ_keys=["compass", "distance"],
                    to_string="distance",
                    space=spaces.Box(low=0, high=np.inf, shape=(), dtype=np.float32),
                )
            )

        super(CompassObservation, self).__init__(handlers=handlers)


class _CompassAngleObservation(KeymapTranslationHandler):
    """
    Handles compass angle observations (converting to the correct angle offset normalized.)
    """

    def __init__(self):
        super().__init__(
            hero_keys=["compassAngle"],
            univ_keys=["compass", "angle"],
            space=spaces.Box(low=-180.0, high=180.0, shape=(), dtype=np.float32),
            to_string="angle",
        )

    def from_universal(self, obs):
        y = np.array(((super().from_universal(obs) * 360.0 + 180) % 360.0) - 180)
        return y



import os
import time
import psutil
import socket
import logging
import threading
from typing import Optional
from contextlib import contextmanager

import Pyro4
import numpy as np

from .instance import MinecraftInstance

logger = logging.getLogger(__name__)


@Pyro4.expose
@Pyro4.behavior(instance_mode="single")
class InstanceManager:
    """The Minecraft instance manager library. The instance manager can be used to allocate and safely terminate
    existing Malmo instances for training agents.

    MineRL Note: This object never needs to be explicitly invoked by the user of the MineRL library as the creation of
    one of the several MineRL environments will automatically query the InstanceManager to create a new instance.

    MineRL Note: In future versions of MineRL the instance manager will become its own daemon process which provides
    instance allocation capability using remote procedure calls.
    """

    MAXINSTANCES = None
    KEEP_ALIVE_PYRO_FREQUENCY = 5
    REMOTE = False

    _instance_pool = []
    _malmo_base_port = 9000
    _jdwp_base_port = 1044  # arbitrary. Used to find available ports for debugging.
    ninstances = 0
    X11_DIR = "/tmp/.X11-unix"
    managed = True

    seed = 42
    rng = np.random.default_rng(seed=seed)

    # this lock allows operating on the instance manager from instances (which
    # run in different worker threads)
    _im_lock: threading.Lock = threading.Lock()

    @classmethod
    def seed_manager(cls, seed: Optional[int]):
        cls.rng = np.random.default_rng(seed=seed)

    @classmethod
    def get_instance(cls, pid, instance_id=None):
        """
        Gets an instance from the instance manager. This method is a context manager
        and therefore when the context is entered the method yields a InstanceManager.Instance
        object which contains the allocated port and host for the given instance that was created.

        Yields:
            The allocated InstanceManager.Instance object.

        Raises:
            RuntimeError: No available instances or the maximum number of allocated instances reached.
            RuntimeError: No available instances and automatic allocation of instances is off.
        """
        if not instance_id:
            # Find an available instance.
            for inst in cls._instance_pool:
                if not inst.locked:
                    inst._acquire_lock(pid)

                    if hasattr(cls, "_pyroDaemon"):
                        cls._pyroDaemon.register(inst)

                    return inst
        # Otherwise make a new instance if possible
        if cls.managed:
            if cls.MAXINSTANCES is None or cls.ninstances < cls.MAXINSTANCES:
                instance_id = cls.ninstances if instance_id is None else instance_id

                cls.ninstances += 1
                # Make the status directory.

                inst = MinecraftInstance(
                    cls._get_valid_port(),
                    instance_id=instance_id,
                    seed=cls.rng.integers(low=0, high=2**31 - 1),
                )

                # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                # Check that not two instances share ports
                # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                dup_ports = [i.uuid for i in cls._instance_pool if i.port == inst.port]
                if len(dup_ports) > 0:
                    # raise exception so we can identify the issue if it happens in experiments
                    raise RuntimeError(
                        f"There are instances with duplicated ports {dup_ports} vs {inst.uuid}"
                    )

                cls._instance_pool.append(inst)
                inst._acquire_lock(pid)

                # find a debugging port for this instance
                if os.getenv("JDWP_ENABLED", False):
                    InstanceManager.set_valid_jdwp_port_for_instance(instance=inst)
                    logger.info(
                        f"Instance {inst.uuid} reserved JDWP port {inst.jdwp_port}."
                    )

                if hasattr(cls, "_pyroDaemon"):
                    cls._pyroDaemon.register(inst)

                return inst

            else:
                raise RuntimeError(
                    "No available instances and max instances reached! :O :O"
                )
        else:
            raise RuntimeError("No available instances and managed flag is off")

    @classmethod
    def shutdown(cls):
        # Iterate over a copy of instance_pool because _stop removes from list
        # This is more time/memory intensive, but allows us to have a modular
        # stop function
        for inst in cls._instance_pool[:]:
            inst.release_lock()
            inst.kill()

    @classmethod
    @contextmanager
    def allocate_pool(cls, num):
        for _ in range(num):
            inst = MinecraftInstance(
                cls._get_valid_port(), seed=cls.rng.integers(low=0, high=2**31 - 1)
            )
            cls._instance_pool.append(inst)
        yield None
        cls.shutdown()

    @classmethod
    def add_existing_instance(cls, port):
        assert cls._is_port_taken(port), "No Malmo mod utilizing the port specified."
        instance = MinecraftInstance(
            port=port, existing=True, seed=cls.rng.integers(low=0, high=2**31 - 1)
        )
        cls._instance_pool.append(instance)
        cls.ninstances += 1
        return instance

    @classmethod
    def add_keep_alive(cls, _pid, _callback):
        logger.debug(
            f"Recieved keep-alive callback from client {_pid}. Starting thread."
        )

        def check_client_connected(client_pid, keep_alive_proxy):
            logger.debug(
                f"Client keep-alive connection monitor started for {client_pid}."
            )
            while True:
                time.sleep(InstanceManager.KEEP_ALIVE_PYRO_FREQUENCY)
                try:
                    keep_alive_proxy.call()
                except:
                    bad_insts = [
                        inst for inst in cls._instance_pool if inst.owner == client_pid
                    ]
                    for inst in bad_insts:
                        inst.close()

        keep_alive_thread = threading.Thread(
            target=check_client_connected, args=(_pid, _callback)
        )
        keep_alive_thread.setDaemon(True)
        keep_alive_thread.start()

    @staticmethod
    def _is_port_taken(port, address="0.0.0.0"):
        if psutil.MACOS or psutil.AIX:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                s.bind((address, port))
                taken = False
            except socket.error as e:
                if e.errno in [98, 10048, 48]:
                    taken = True
                else:
                    raise e

            s.close()
            return taken
        else:
            ports = [x.laddr.port for x in psutil.net_connections()]
            return port in ports

    @classmethod
    def _port_in_instance_pool(cls, port):
        # Ideally, this should be covered by other cases, but there may be delay
        # in when the ports get "used"
        for instance in cls._instance_pool:
            if port == instance.port or port == instance.jdwp_port:
                return True
        return False

    @classmethod
    def configure_malmo_base_port(cls, malmo_base_port):
        """Configure the lowest or base port for Malmo"""
        cls._malmo_base_port = malmo_base_port

    @classmethod
    def _get_valid_port(cls):
        malmo_base_port = cls._malmo_base_port
        port = (cls.ninstances % 5000) + malmo_base_port
        port += (17 * os.getpid()) % 3989
        while cls._is_port_taken(port) or cls._port_in_instance_pool(port):
            port += 1
        return port

    @classmethod
    def set_valid_jdwp_port_for_instance(cls, instance) -> None:
        """
        Find a valid port for JDWP (Java Debug Wire Protocol), so that the instance can be debugged
        with an attached debugger. The port is set in the instance, so that other instances can
        check whether the port is reserved.
        :param instance: Instance to find and port for, and where we will set the jdwp port.
        """
        # since we need to check whether other instances have ports claimed, we should find ports
        # for instances already in the pool
        assert (
            instance in cls._instance_pool
        ), "Attempted to find jdwp port for instance not in the pool."

        port = cls._jdwp_base_port
        last_port_to_check = port + 256  # do not try forever

        # this needs to be atomic, otherwise other threads checking for ports might grab the same
        # port
        with cls._im_lock:

            # find a port
            while cls._is_port_taken(port) or cls._port_in_instance_pool(port):
                port += 1
                if port >= last_port_to_check:
                    instance.jdwp_port = None
                    break

            # set the port in the instance before releasing the lock
            instance.jdwp_port = port

    @classmethod
    def is_remote(cls):
        return cls.REMOTE



"""
Execute MineCraft native commands
"""
import logging
from typing import Optional, List, Union

import numpy as np

from .inventory import (
    InventoryItem,
    parse_inventory_item,
    map_slot_number_to_cmd_slot,
)


class CMDExecutor:
    logger = logging.getLogger("cmd_executor")
    valid_cmds = {
        "summon",
        "kill",
        "time",
        "weather",
        "replaceitem",
        "tp",
        "clear",
        "setblock",
        "spreadplayers",
    }

    def __init__(self, world, raise_error_on_invalid_cmds: bool = False):
        # avoid circular import
        from .sim import MineDojoSim

        self._world: MineDojoSim = world
        self._raise_error_on_invalid_cmds = raise_error_on_invalid_cmds
        self._world_action_space = world.action_space

    def execute_cmd(self, cmd: str, action: Optional[dict] = None):
        if (not cmd.startswith("/")) or (cmd.split()[0][1:] not in self.valid_cmds):
            if not self._raise_error_on_invalid_cmds:
                self.logger.warning(f"Invalid cmd {cmd}, skipping...")
                if action is None:
                    return (
                        self._world.prev_obs,
                        0,
                        self._world.is_terminated,
                        self._world.prev_info,
                    )
                else:
                    return self._world.step(action)
            else:
                raise ValueError(f"Invalid cmd {cmd}")
        else:
            action = action or self._world.action_space.no_op()
            action["chat"] = cmd
            return self._world.step(action)

    def spawn_mobs(
        self,
        mobs: Union[str, List[str]],
        rel_positions: Union[np.ndarray, list],
        action: Optional[dict] = None,
    ):
        if isinstance(mobs, str):
            mobs = [mobs]
        if isinstance(rel_positions, list):
            rel_positions = np.array(rel_positions)
        if rel_positions.ndim == 1:
            rel_positions = rel_positions[np.newaxis, ...]
        if len(mobs) > 1 and len(rel_positions) == 1:
            rel_positions = np.repeat(rel_positions, len(mobs), axis=0)
        assert (
            rel_positions.ndim == 2
        ), f"Expect `rel_positions` to be 2 dimensional, but got {rel_positions.ndim} dims"
        assert (
            rel_positions.shape[1] == 3
        ), "Expect `rel_positions` to contain x, y, and z"
        assert len(mobs) == len(
            rel_positions
        ), f"Expect {len(mobs)} relative positions, but got {len(rel_positions)}"

        obs, info = self._world.prev_obs, self._world.prev_info
        for mob, rel_pos in zip(mobs, rel_positions):
            cmd = f"/summon {mob} ~{int(rel_pos[0])} ~{int(rel_pos[1])} ~{int(rel_pos[2])}"
            obs, _, _, info = self.execute_cmd(cmd, action)
        return obs, 0, self._world.is_terminated, info

    def set_block(
        self,
        blocks: Union[str, List[str]],
        rel_positions: Union[np.ndarray, list],
        action: Optional[dict] = None,
    ):
        if isinstance(blocks, str):
            blocks = [blocks]
        if isinstance(rel_positions, list):
            rel_positions = np.array(rel_positions)
        if rel_positions.ndim == 1:
            rel_positions = rel_positions[np.newaxis, ...]
        if len(blocks) > 1 and len(rel_positions) == 1:
            rel_positions = np.repeat(rel_positions, len(blocks), axis=0)
        assert (
            rel_positions.ndim == 2
        ), f"Expect `rel_positions` to be 2 dimensional, but got {rel_positions.ndim} dims"
        assert (
            rel_positions.shape[1] == 3
        ), "Expect `rel_positions` to contain x, y, and z"
        assert len(blocks) == len(
            rel_positions
        ), f"Expect {len(blocks)} relative positions, but got {len(rel_positions)}"

        obs, info = self._world.prev_obs, self._world.prev_info
        for block, rel_pos in zip(blocks, rel_positions):
            cmd = f"/setblock ~{int(rel_pos[0])} ~{int(rel_pos[1])} ~{int(rel_pos[2])} {block}"
            obs, _, _, info = self.execute_cmd(cmd, action)
        return obs, 0, self._world.is_terminated, info

    def clear_inventory(self, action: Optional[dict] = None):
        obs, _, _, info = self.execute_cmd("/clear", action)
        return obs, 0, self._world.is_terminated, info

    def set_inventory(
        self, inventory_list: List[InventoryItem], action: Optional[dict] = None
    ):
        obs, info = self._world.prev_obs, self._world.prev_info
        for inventory_item in inventory_list:
            slot, item_dict = parse_inventory_item(inventory_item)
            obs, _, _, info = self.execute_cmd(
                f'/replaceitem entity @p {map_slot_number_to_cmd_slot(slot)} minecraft:{item_dict["type"]} {item_dict["quantity"]} {item_dict["metadata"]}',
                action,
            )
        return obs, 0, self._world.is_terminated, info

    def teleport_agent(self, x, y, z, yaw, pitch, action: Optional[dict] = None):
        obs, _, _, info = self.execute_cmd(f"/tp {x} {y} {z} {yaw} {pitch}", action)
        return obs, 0, self._world.is_terminated, info

    def kill_agent(self, action: Optional[dict] = None):
        obs, _, _, info = self.execute_cmd("/kill", action)
        return obs, 0, self._world.is_terminated, info

    def set_time(self, time: int, action: Optional[dict] = None):
        obs, _, _, info = self.execute_cmd(f"/time set {time}", action)
        return obs, 0, self._world.is_terminated, info

    def set_weather(self, weather: str, action: Optional[dict] = None):
        obs, _, _, info = self.execute_cmd(f"/weather {weather}", action)
        return obs, 0, self._world.is_terminated, info

    def random_teleport(self, max_range: int, action: Optional[dict] = None):
        obs, _, _, info = self.execute_cmd(
            f"/spreadplayers ~ ~ 0 {max_range} false @p", action
        )
        return obs, 0, self._world.is_terminated, info



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
import logging
from typing import Any, Dict, List

import numpy as np

from .. import spaces as spaces
from ..handler import Handler


class TranslationHandler(Handler):
    """
    An agent handler to be added to the mission XML.
    This is useful as it defines basically all of the interfaces
    between universal action format, hero (malmo), and herobriane (ML stuff).
    """

    def __init__(self, space: spaces.MineRLSpace, **other_kwargs):
        self.space = space

    def from_hero(self, x: Dict[str, Any]):
        """
        Converts a "hero" representation of an instance of this handler
        to a member of the space.
        """
        raise NotImplementedError()

    def to_hero(self, x) -> str:
        """
        Takes an instance of the handler, x in self.space, and maps it to
        the "hero" representation thereof.
        """
        raise NotImplementedError()

    def from_universal(self, x: Dict[str, Any]):
        """
        Converts a universal representation of the handler (e.g. universal action/observation)
        """
        raise NotImplementedError()


# TODO: ONLY WORKS FOR OBSERVATIONS.
# TODO: Consider moving this to observations.
class KeymapTranslationHandler(TranslationHandler):
    def __init__(
        self,
        hero_keys: List[str],
        univ_keys: List[str],
        space: spaces.MineRLSpace,
        default_if_missing=None,
        to_string: str = None,
        ignore_missing: bool = False,
    ):
        """
        Wrapper for simple observations which just remaps keys.
        :param keys: list of nested dictionary keys from the root of the observation dict
        :param space: gym space corresponding to the shape of the returned value
        :param default_if_missing: value for handler to take if missing in the observation dict
        :param ignore_missing: If we should throw a warning when corresponding json field is
            missing from the observation.
        """
        super().__init__(space)
        self._to_string = to_string if to_string else hero_keys[-1]
        self.hero_keys = hero_keys
        self.univ_keys = univ_keys
        self.default_if_missing = default_if_missing
        self.ignore_missing = ignore_missing

    @property
    def logger(self):
        return logging.getLogger(f"{__name__}.{self.to_string()}")

    def walk_dict(self, d, keys, dtype=None):
        for key in keys:
            if key in d:
                d = d[key]
            else:
                if self.default_if_missing is not None:
                    if not self.ignore_missing:
                        self.logger.error(
                            f"No {keys[-1]} observation! Yielding default value "
                            f"{self.default_if_missing} for {'/'.join(keys)}"
                        )
                    return np.array(self.default_if_missing, dtype=dtype)
                else:
                    raise KeyError()
        return np.array(d, dtype=dtype)

    def to_hero(self, x) -> str:
        """What does it mean to do a keymap translation here?
        Since hero sends things as commands perhaps we could generalize
        this beyond observations.
        """
        raise NotImplementedError()

    def from_hero(self, hero_dict, dtype=None):
        return self.walk_dict(hero_dict, self.hero_keys, dtype=dtype)

    def from_universal(self, univ_dict):
        return self.walk_dict(univ_dict, self.univ_keys)

    def to_string(self) -> str:
        return self._to_string


class TranslationHandlerGroup(TranslationHandler):
    """Combines several space handlers into a single handler group."""

    def __init__(self, handlers: List[TranslationHandler]):
        self.handlers = sorted(handlers, key=lambda x: x.to_string())
        super(TranslationHandlerGroup, self).__init__(
            spaces.Dict([(h.to_string(), h.space) for h in self.handlers])
        )

    def to_hero(self, x: Dict[str, Any]) -> str:
        """Produces a string from an object X contained in self.space
        into a string by calling all of the corresponding
        to_hero methods and joining them with new lines
        """

        return "\n".join(
            [self.handler_dict[s].to_hero(x[s]) for s in self.handler_dict]
        )

    def from_hero(self, x: Dict[str, Any]) -> Dict[str, Any]:
        """Applies the constituent from_hero methods on the object X
        and builds a dictionary with keys corresponding to the constituent
        handlers applied."""

        return {h.to_string(): h.from_hero(x) for h in self.handlers}

    def from_universal(self, x: Dict[str, Any]) -> Dict[str, Any]:
        """Performs the same operation as from_hero except with from_universal."""
        return {h.to_string(): h.from_universal(x) for h in self.handlers}

    @property
    def handler_dict(self) -> Dict[str, Handler]:
        return {h.to_string(): h for h in self.handlers}



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
import numpy as np

import minedojo.sim.spaces as spaces
from minedojo.sim.handlers.agent.action import Action


class CameraAction(Action):
    """
    Uses <delta_pitch, delta_yaw> vector in degrees to rotate the camera. pitch range [-180, 180], yaw range [-180, 180]
    """

    def to_string(self):
        return "camera"

    def xml_template(self) -> str:
        return str("<CameraCommands/>")

    def __init__(self):
        # TODO (minerl): Document and clean this wierd _ magic.
        self._command = "camera"
        super().__init__(
            self.command,
            spaces.Box(
                low=-180,
                high=180,
                shape=(2,),
                dtype=float,
            ),
        )



from typing import Union, Optional

import gym

from .ar_masks_wrapper import ARMasksWrapper as _ARMasksWrapper
from .nn_action_space_wrapper import NNActionSpaceWrapper as _NNActionSpaceWrapper
from .delta_inventory_wrapper import DeltaInventoryWrapper as _DeltaInventoryObsWrapper


class ARNNWrapper(gym.Wrapper):
    def __init__(
        self,
        sim,
        cam_interval: Union[int, float] = 15,
        strict_check: bool = True,
        op_action_idx: int = 5,
        craft_action_idx: int = 4,
        craft_arg_idx: int = 6,
        n_increased: int = 1,
        n_decreased: int = 4,
        default_item_name: str = "air",
        action_categories_and_num_args: Optional[dict[str, int]] = None,
    ):
        sim = _DeltaInventoryObsWrapper(
            _ARMasksWrapper(
                _NNActionSpaceWrapper(
                    env=sim,
                    discretized_camera_interval=cam_interval,
                    strict_check=strict_check,
                ),
                action_categories_and_num_args=action_categories_and_num_args,
            ),
            op_action_idx=op_action_idx,
            craft_action_idx=craft_action_idx,
            craft_arg_idx=craft_arg_idx,
            n_increased=n_increased,
            n_decreased=n_decreased,
            default_item_name=default_item_name,
        )
        self.cam_interval = cam_interval
        super().__init__(env=sim)

    def reverse_action(self, action):
        return self.env.env.env.reverse_action(action)



from __future__ import annotations

from copy import deepcopy
from typing import List, Dict, Any, Union, Tuple, Optional

import gym
import numpy as np

from ...sim import MineDojoSim
from ...sim.wrappers import FastResetWrapper
from .utils import (
    check_success_base,
    reward_fn_base,
    extra_spawn_condition_base,
    always_satisfy_condition,
)


__all__ = ["MetaTaskBase", "ExtraSpawnMetaTaskBase"]


class MetaTaskBase(gym.Wrapper):
    """
    Base class for all meta tasks.

    Args:
        fast_reset: If ``True``, reset using MC native command `/kill`, instead of waiting for re-generating new worlds.
            Default: ``True``.

            .. warning::
                Side effects:

                1. Changes to the world will not be reset. E.g., if the agent chops lots of trees then calling
                fast reset will not restore those trees.

                2. If you specify agent starting health and food, these specs will only be respected at the first reset
                (i.e., generating a new world) but will not be respected in the following resets (i.e., reset using MC cmds).
                So be careful to use this wrapper if your usages require specific initial health/food.

                3. Statistics/achievements will not be reset. This wrapper will maintain a property ``info_prev_reset``.
                If your tasks use stat/achievements to evaluation, please retrieve this property and compute differences.

        reward_fns: The reward functions of the task.
        success_criteria: The success criteria of the task.
    """

    _prompt_template: str
    _use_specified_prompt: bool = False
    _specified_prompt: str | None = None
    _guidance: str | None = None

    def __init__(
        self,
        *,
        fast_reset: bool = True,
        fast_reset_random_teleport_range: Optional[int] = None,
        success_criteria: List[check_success_base],
        reward_fns: List[reward_fn_base],
        **kwargs,
    ):
        sim = MineDojoSim(**kwargs)
        self._fast_reset = fast_reset
        if fast_reset:
            sim = FastResetWrapper(
                sim, random_teleport_range=fast_reset_random_teleport_range
            )
        super().__init__(env=sim)

        self._ini_info_dict = None
        self._pre_info_dict = None
        self._elapsed_timesteps = None
        self._is_successful: bool = False

        self._success_criteria = success_criteria
        self._reward_fns = reward_fns

    @property
    def is_successful(self):
        return self._is_successful

    def reset(self):
        """Resets the environment to an initial state and returns an initial observation.

        Return:
            Agent’s initial observation.
        """
        obs = self.env.reset()
        info = self.env.prev_info
        obs, info = self._after_sim_reset_hook(obs, info)
        self._ini_info_dict = (
            self.env.info_prev_reset or info if self._fast_reset else info
        )
        self._pre_info_dict = deepcopy(info)
        self._elapsed_timesteps = 0
        self._is_successful = False
        return obs

    def step(self, action):
        """Run one timestep of the environment’s dynamics. Accepts an action and returns next_obs, reward, done, info.

        Args:
            action: The action of the agent in current step.

        Return:
            A tuple (obs, reward, done, info)
            - ``dict`` - Agent’s next observation.
            - ``float`` - Amount of reward returned after executing previous action.
            - ``bool`` - Whether the episode has ended.
            - ``dict`` - Contains auxiliary diagnostic information (helpful for debugging, and sometimes learning).
        """
        obs, _, _, info = self.env.step(action)
        self._elapsed_timesteps += 1
        reward = self._compute_reward_hook(
            ini_info=self._ini_info_dict,
            pre_info=self._pre_info_dict,
            cur_info=info,
            elapsed_timesteps=self._elapsed_timesteps,
        )
        self._is_successful = self._determine_success_hook(
            ini_info=self._ini_info_dict,
            cur_info=info,
            elapsed_timesteps=self._elapsed_timesteps,
        )
        done = self.env.is_terminated or self._is_successful
        self._pre_info_dict = deepcopy(info)
        return obs, reward, done, info

    def get_prompt(self, **kwargs) -> str:
        """Get the prompt of the task"""
        return (
            self._specified_prompt
            if self._use_specified_prompt
            else self._prompt_template.format(**kwargs)
        )

    def specify_prompt(self, prompt: str):
        """Allow users to customize the prompt"""
        self._specified_prompt = prompt
        self._use_specified_prompt = True

    @property
    def task_guidance(self):
        return self._guidance

    def specify_guidance(self, guidance: str):
        """Allow users to customize the guidance"""
        self._guidance = guidance

    def _after_sim_reset_hook(
        self, reset_obs: Dict[str, Any], reset_info: Dict[str, Any]
    ) -> (Dict[str, Any], Dict[str, Any]):
        """
        Called immediately after `self.env.reset()` in `self.reset()`.
        Override to implement initial conditions, e.g., initial mob spawn, initial teleporting, etc.
        Takes input of obs and info from sim reset.
        Should return resulting obs and info.
        """
        return reset_obs, reset_info

    def _compute_reward_hook(
        self,
        ini_info: Dict[str, Any],
        pre_info: Dict[str, Any],
        cur_info: Dict[str, Any],
        elapsed_timesteps: int,
    ) -> Union[int, float]:
        return sum(
            [
                reward_fn(
                    ini_info_dict=ini_info,
                    pre_info_dict=pre_info,
                    cur_info_dict=cur_info,
                    elapsed_timesteps=elapsed_timesteps,
                )
                for reward_fn in self._reward_fns
            ]
        )

    def _determine_success_hook(
        self, ini_info: Dict[str, Any], cur_info: Dict[str, Any], elapsed_timesteps: int
    ) -> bool:
        return all(
            [
                check_success(
                    ini_info_dict=ini_info,
                    cur_info_dict=cur_info,
                    elapsed_timesteps=elapsed_timesteps,
                )
                for check_success in self._success_criteria
            ]
        )


class ExtraSpawnMetaTaskBase(MetaTaskBase):
    """
    Base class for all meta tasks that need extra spawn rate for resources or mobs.

    Args:
        extra_spawn_condition: The extra spawn rate applies only when the extra spawn conditions are satisfied.

        extra_spawn_range_high: The upper bound on each horizontal axis from the center of the area to extra spawn.

        extra_spawn_range_low: The lower bound on each horizontal axis from the center of the area to extra spawn.

        extra_spawn_rate: The probability of extra spawning in each step.
                Default: ``None``.

        fast_reset: If ``True``, fast reset will be used in the task. See ``fast_reset`` in ``MetaTaskBase``.
                Default: ``True``.

        initial_mobs: The types of mobs that are spawned initially.
                Default: ``None``.

        initial_mob_spawn_range_high: The upper bound on each horizontal axis from the center of the area to spawn initially.
                Default: ``None``.

        initial_mob_spawn_range_low: The lower bound on each horizontal axis from the center of the area to spawn initially.
                Default: ``None``.

        reward_fns: The reward functions of the task.

        success_criteria: The success criteria of the task.
    """

    # TODO (yunfan): complete these lists
    by_setblock = ["diamond_ore", "gold_ore", "iron_ore", "coal_ore"]
    by_summon = ["pig", "cow", "bat", "cat", "chicken", "horse", "sheep", "zombie"]

    def __init__(
        self,
        *,
        initial_mobs: Optional[Union[str, List[str]]] = None,
        initial_mob_spawn_range_low: Optional[Tuple[int, int, int]] = None,
        initial_mob_spawn_range_high: Optional[Tuple[int, int, int]] = None,
        extra_spawn_rate: Optional[Dict[str, float]] = None,
        extra_spawn_condition: Optional[
            Union[extra_spawn_condition_base, Dict[str, extra_spawn_condition_base]]
        ],
        extra_spawn_range_low: Optional[Tuple[int, int, int]],
        extra_spawn_range_high: Optional[Tuple[int, int, int]],
        fast_reset: bool = True,
        fast_reset_random_teleport_range: Optional[int] = None,
        success_criteria: List[check_success_base],
        reward_fns: List[reward_fn_base],
        **kwargs,
    ):
        """
        base task class for meta tasks that require extra spawn, e.g., harvest, combat, and techtree
        """
        self._extra_spawn_rate = None
        if extra_spawn_rate is not None:
            assert all(
                [1.0 >= rate >= 0.0 for rate in extra_spawn_rate.values()]
            ), "extra spawn rate must <= 1.0 and >= 0.0"
            assert all(
                [
                    name in self.by_summon or name in self.by_setblock
                    for name in extra_spawn_rate.keys()
                ]
            ), f"{extra_spawn_rate.keys()} should belong to either {self.by_summony} or {self.by_setblock}"
            if extra_spawn_condition is None:
                extra_spawn_condition = {
                    k: always_satisfy_condition for k in extra_spawn_rate.keys()
                }
            else:
                if isinstance(extra_spawn_condition, dict):
                    assert set(extra_spawn_rate.keys()) == set(
                        extra_spawn_condition.keys()
                    ), (
                        f"extra_spawn_rate must match extra_spawn_condition, "
                        f"but got {extra_spawn_rate.keys()} and {extra_spawn_condition.keys()}"
                    )
                else:
                    extra_spawn_condition = {
                        k: extra_spawn_condition for k in extra_spawn_rate.keys()
                    }
            self._extra_spawn_rate = extra_spawn_rate
            self._extra_spawn_rates_and_conditions = {
                k: (extra_spawn_rate[k], extra_spawn_condition[k])
                for k in extra_spawn_rate.keys()
            }
            assert (
                extra_spawn_range_high is not None and extra_spawn_range_low is not None
            )
            assert len(extra_spawn_range_high) == 3 and len(extra_spawn_range_low) == 3
            low = np.repeat(
                np.array(extra_spawn_range_low)[np.newaxis, ...],
                len(extra_spawn_rate),
                axis=0,
            )
            high = np.repeat(
                np.array(extra_spawn_range_high)[np.newaxis, ...],
                len(extra_spawn_rate),
                axis=0,
            )
            self._extra_spawn_range_space = gym.spaces.Box(
                low=low, high=high, seed=kwargs["seed"]
            )
            self._rng = np.random.default_rng(seed=kwargs["seed"])

        super().__init__(
            fast_reset=fast_reset,
            fast_reset_random_teleport_range=fast_reset_random_teleport_range,
            success_criteria=success_criteria,
            reward_fns=reward_fns,
            **kwargs,
        )

        initial_mobs = initial_mobs or []
        if isinstance(initial_mobs, str):
            initial_mobs = [initial_mobs]
        self._initial_mobs = initial_mobs
        if len(initial_mobs) > 0:
            assert len(initial_mob_spawn_range_low) == 3
            assert len(initial_mob_spawn_range_high) == 3
            low = np.repeat(
                np.array(initial_mob_spawn_range_low)[np.newaxis, ...],
                len(initial_mobs),
                axis=0,
            )
            high = np.repeat(
                np.array(initial_mob_spawn_range_high)[np.newaxis, ...],
                len(initial_mobs),
                axis=0,
            )
            self._mob_spawn_range_space = gym.spaces.Box(
                low=low, high=high, seed=kwargs["seed"]
            )

    def step(self, action):
        if self._extra_spawn_rate is None:
            return super().step(action=action)
        else:
            rel_positions = self._extra_spawn_range_space.sample()
            for (name, (rate, condition)), pos in zip(
                self._extra_spawn_rates_and_conditions.items(), rel_positions
            ):
                if self._rng.random() <= rate and condition(
                    ini_info_dict=self._ini_info_dict,
                    pre_info_dict=self._pre_info_dict,
                    elapsed_timesteps=self._elapsed_timesteps,
                ):
                    if name in self.by_summon:
                        obs, _, _, info = self.env.spawn_mobs(name, pos)
                    elif name in self.by_setblock:
                        obs, _, _, info = self.env.set_block(name, pos)
            return super().step(action=action)

    def _after_sim_reset_hook(
        self, reset_obs: Dict[str, Any], reset_info: Dict[str, Any]
    ) -> (Dict[str, Any], Dict[str, Any]):
        obs, info = reset_obs, reset_info
        if len(self._initial_mobs) > 0:
            mobs_rel_positions = self._mob_spawn_range_space.sample()
            obs, _, _, info = self.env.spawn_mobs(
                self._initial_mobs, mobs_rel_positions
            )
        return obs, info



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton

"""
World handlers provide a number of ways to generate and modify the Minecraft world
(e.g. specifying the type of world to be created, like Superflat, or drawing shapes and blocks in the world).

When used to create a Gym environment, they should be passed to :code:`create_server_world_generators`
"""

from typing import Union
from minedojo.sim.mc_meta import mc
from minedojo.sim.handler import Handler


class DefaultWorldGenerator(Handler):
    """Generates a world using minecraft procedural generation (this is the default world type in minecraft).

    Args:
        force_reset (bool, optional): If the world should be reset every episode.. Defaults to True.
        world_seed: https://minecraft.fandom.com/wiki/Seed_(level_generation)

    Example usage:

    .. code-block:: python

        # Generates a default world that does not reset every episode (e.g. if blocks get broken in one episode
        # they will not be replaced in the next)
        DefaultWorldGenerator(False, "")
    """

    def to_string(self) -> str:
        return "default_world_generator"

    def xml_template(self) -> str:
        return str(
            """<DefaultWorldGenerator
                forceReset="{{force_reset | string | lower}}"
                seed="{{world_seed}}"
                generatorOptions="{}"/>
            """
        )

    def __init__(self, force_reset=True, world_seed: str = ""):
        self.force_reset = force_reset
        self.world_seed = world_seed


class FileWorldGenerator(Handler):
    """Generates a world from a file."""

    def to_string(self) -> str:
        return "file_world_generator"

    def xml_template(self) -> str:
        return str(
            """<FileWorldGenerator
                destroyAfterUse = "{{destroy_after_use | string | lower}}"
                src = "{{filename}}" />
            """
        )

    def __init__(self, filename: str, destroy_after_use: bool = True):
        self.filename = filename
        self.destroy_after_use = destroy_after_use


#  <FlatWorldGenerator forceReset="true"/>
class FlatWorldGenerator(Handler):
    """
    Generates a world that is a flat landscape.

    Example usage:

    .. code-block:: python
        # Create a superflat world with layers as follow:
        # 1 layer of grass blocks above 2 layers of dirt above 1 layer of bedrock
        # You can use websites like "`Minecraft Tools`_" to easily customize superflat world layers.
        FlatWorldGenerator(generatorString="1;7,2x3,2;1")

    .. _Minecraft Tools: https://minecraft.tools/en/flat.php?biome=1&bloc_1_nb=1&bloc_1_id=2&bloc_2_nb=2&bloc_2_id=3%2F00&bloc_3_nb=1&bloc_3_id=7&village_size=1&village_distance=32&mineshaft_chance=1&stronghold_count=3&stronghold_distance=32&stronghold_spread=3&oceanmonument_spacing=32&oceanmonument_separation=5&biome_1_distance=32&valid=Create+the+Preset#seed
    """

    def to_string(self) -> str:
        return "flat_world_generator"

    def xml_template(self) -> str:
        return str(
            """<FlatWorldGenerator
                forceReset="{{force_reset | string | lower}}"
                generatorString="{{generatorString}}" />
            """
        )

    def __init__(self, force_reset: bool = True, generatorString: str = ""):
        self.force_reset = force_reset
        self.generatorString = generatorString


#  <BiomeGenerator forceReset="true" biome="3"/>
class BiomeGenerator(Handler):
    def to_string(self) -> str:
        return "biome_generator"

    def xml_template(self) -> str:
        return str(
            """<BiomeGenerator
                forceReset="{{force_reset | string | lower}}"
                seed="{{world_seed}}"
                biome="{{biome_id}}" />
            """
        )

    def __init__(
        self, biome: Union[int, str], force_reset: bool = True, world_seed: str = ""
    ):
        if isinstance(biome, str):
            biome = mc.BIOMES_MAP[biome]
        self.biome_id = biome
        self.force_reset = force_reset
        self.world_seed = world_seed


class DrawingDecorator(Handler):
    """
    Draws shapes (e.g. spheres, cuboids) in the world.

    Example usage:

    .. code-block:: python

        # draws an empty square of gold blocks
        DrawingDecorator('
            <DrawCuboid x1="3" y1="4" z1="3" x2="3" y2="6" z2="-3" type="gold_block"/>
            <DrawCuboid x1="3" y1="4" z1="3" x2="-3" y2="6" z2="3" type="gold_block"/>
            <DrawCuboid x1="-3" y1="4" z1="-3" x2="3" y2="6" z2="-3" type="gold_block"/>
            <DrawCuboid x1="-3" y1="4" z1="-3" x2="-3" y2="6" z2="3" type="gold_block"/>
        ')

    See Project Malmo for more
    """

    def __init__(self, to_draw: str):
        self.to_draw = to_draw

    def xml_template(self) -> str:
        tmp = """<DrawingDecorator>{{to_draw}}</DrawingDecorator>"""
        return tmp

    def to_string(self) -> str:
        return "drawing_decorator"


class VillageSpawnDecorator(Handler):
    def xml_template(self) -> str:
        tmp = """<VillageSpawnDecorator />"""
        return tmp

    def to_string(self) -> str:
        return "village_spawn"



import numpy as np

from minedojo.sim import spaces
from minedojo.sim.handlers.translation import KeymapTranslationHandler


class VoxelObservation(KeymapTranslationHandler):
    """
    Handles voxel observations.
    Returned voxels are in (x, y, z) order, where x, y, z are all in ascending order.
    """

    def to_hero(self, x) -> str:
        pass

    def to_string(self):
        return "voxels"

    def xml_template(self) -> str:
        return str(
            """
            <ObservationFromGrid>                      
                <Grid name="voxels">                        
                    <min x="{{xmin}}" y="{{ymin}}" z="{{zmin}}"/>                        
                    <max x="{{xmax}}" y="{{ymax}}" z="{{zmax}}"/>                      
                </Grid>                  
            </ObservationFromGrid>"""
        )

    def __init__(self, limits=((-3, 3), (-1, 3), (-3, 3))):
        self.xmin = limits[0][0]
        self.ymin = limits[1][0]
        self.zmin = limits[2][0]
        self.xmax = limits[0][1]
        self.ymax = limits[1][1]
        self.zmax = limits[2][1]
        self.grid_size = [1 + b - a for a, b in limits]

        space = spaces.Dict(
            {
                "block_name": spaces.Text(shape=self.grid_size),
                # max block meta is 120, i.e., item id 383:120 spawn egg for villager
                # see https://minecraft-ids.grahamedgecombe.com/
                "block_meta": spaces.Box(
                    low=0, high=120, shape=self.grid_size, dtype=np.int64
                ),
                "is_collidable": spaces.Box(
                    low=0, high=1, shape=self.grid_size, dtype=bool
                ),
                "is_tool_not_required": spaces.Box(
                    low=0, high=1, shape=self.grid_size, dtype=bool
                ),
                "blocks_movement": spaces.Box(
                    low=0, high=1, shape=self.grid_size, dtype=bool
                ),
                "is_liquid": spaces.Box(
                    low=0, high=1, shape=self.grid_size, dtype=bool
                ),
                "is_solid": spaces.Box(low=0, high=1, shape=self.grid_size, dtype=bool),
                "can_burn": spaces.Box(low=0, high=1, shape=self.grid_size, dtype=bool),
                "blocks_light": spaces.Box(
                    low=0, high=1, shape=self.grid_size, dtype=bool
                ),
                "cos_look_vec_angle": spaces.Box(
                    low=-1, high=1, shape=self.grid_size, dtype=np.float32
                ),
            }
        )
        # gym.space.Dict will sort keys! So we keep the order by ourselves
        self._key_list = [
            "block_name",
            "block_meta",
            "is_collidable",
            "is_tool_not_required",
            "blocks_movement",
            "is_liquid",
            "is_solid",
            "can_burn",
            "blocks_light",
            "cos_look_vec_angle",
        ]
        super().__init__(hero_keys=["voxels"], univ_keys=["voxels"], space=space)

    def from_hero(self, obs):
        voxels_arr = obs[self.hero_keys[0]]
        assert len(voxels_arr) == np.prod(self.grid_size) * len(
            self._key_list
        ), "INTERNAL"
        # yunfan: note that returns from java side are in F order, we need to use F order to be consistent
        return {
            key: np.array(
                [
                    voxels_arr[i]
                    for i in range(bias, len(voxels_arr), len(self._key_list))
                ],
                dtype=self.space[key].dtype,
            ).reshape(self.space[key].shape, order="F")
            for bias, key in enumerate(self._key_list)
        }

    def __or__(self, other):
        """
        Combines two voxel observations into one. If all of the properties match return self
        otherwise raise an exception.
        """
        if (
            isinstance(other, VoxelObservation)
            and self.grid_min == other.grid_min
            and self.grid_max == other.grid_max
        ):
            return self
        else:
            raise ValueError("Incompatible observables!")


class RichLidarObservation(KeymapTranslationHandler):
    """
    Handles rich LIDAR observations.
    """

    def to_hero(self, x) -> str:
        pass

    def to_string(self):
        return "rays"

    def xml_template(self) -> str:
        return str(
            """
            <ObservationFromRichLidar>
                {% for ray in rays %}
                    <RayOffset pitch="{{ray[0]}}" yaw="{{ray[1]}}" distance="{{ray[2]}}"/>
                {% endfor %}
            </ObservationFromRichLidar>
        """
        )

    def __init__(self, rays=None):
        # Note rays use [pitch, yaw, distance]:
        # The pitch (in radians) is relative to lookVec
        # The yaw (in radians) is relative to lookVec
        # The distance (in meters) is the maximum distance for the ray from eyePos

        if rays is None:
            rays = [
                (0.0, 0.0, 10.0),
            ]
        self.rays = rays
        self.num_rays = len(rays)

        _shape = [self.num_rays]
        space = spaces.Dict(
            {
                "block_name": spaces.Text(shape=_shape),
                "block_distance": spaces.Box(
                    low=-1, high=np.inf, shape=_shape, dtype=np.float32
                ),
                # max block meta is 120, i.e., item id 383:120 spawn egg for villager
                # see https://minecraft-ids.grahamedgecombe.com/
                "block_meta": spaces.Box(low=0, high=120, shape=_shape, dtype=np.int64),
                "harvest_level": spaces.Box(
                    low=-1, high=4, shape=_shape, dtype=np.int64
                ),
                "is_tool_not_required": spaces.Box(
                    low=0, high=1, shape=_shape, dtype=bool
                ),
                "blocks_movement": spaces.Box(low=0, high=1, shape=_shape, dtype=bool),
                "is_liquid": spaces.Box(low=0, high=1, shape=_shape, dtype=bool),
                "is_solid": spaces.Box(low=0, high=1, shape=_shape, dtype=bool),
                "can_burn": spaces.Box(low=0, high=1, shape=_shape, dtype=bool),
                "entity_name": spaces.Text(shape=_shape),
                "entity_distance": spaces.Box(
                    low=-1, high=np.inf, shape=_shape, dtype=np.float32
                ),
                "ray_pitch": spaces.Box(
                    low=-np.inf, high=np.inf, shape=_shape, dtype=np.float32
                ),
                "ray_yaw": spaces.Box(
                    low=-np.inf, high=np.inf, shape=_shape, dtype=np.float32
                ),
                "traced_block_x": spaces.Box(
                    low=-np.inf, high=np.inf, shape=_shape, dtype=np.float32
                ),
                "traced_block_y": spaces.Box(
                    low=-np.inf, high=np.inf, shape=_shape, dtype=np.float32
                ),
                "traced_block_z": spaces.Box(
                    low=-np.inf, high=np.inf, shape=_shape, dtype=np.float32
                ),
            }
        )
        # gym.space.Dict will sort keys! So we keep the order by ourselves
        self._key_list = [
            "block_name",
            "block_distance",
            "block_meta",
            "harvest_level",
            "is_tool_not_required",
            "blocks_movement",
            "is_liquid",
            "is_solid",
            "can_burn",
            "entity_name",
            "entity_distance",
            "ray_pitch",
            "ray_yaw",
            "traced_block_x",
            "traced_block_y",
            "traced_block_z",
        ]

        super().__init__(
            hero_keys=["rays"],
            univ_keys=["rays"],
            space=space,
        )

    def from_hero(self, obs):
        raytrace_arr = obs[self.hero_keys[0]]
        assert len(raytrace_arr) == self.num_rays * len(self._key_list), "INTERNAL"
        return {
            key: np.array(
                [
                    self._map_entity_name(raytrace_arr[i])
                    if key == "entity_name"
                    else raytrace_arr[i]
                    for i in range(bias, len(raytrace_arr), len(self._key_list))
                ],
                dtype=self.space[key].dtype,
            ).reshape(self.space[key].shape)
            for bias, key in enumerate(self._key_list)
        }

    def __or__(self, other):
        """
        Combines two rich lidar observations into one.
        """
        if isinstance(other, RichLidarObservation):
            all_rays = self.rays + other.rays
            return RichLidarObservation(rays=all_rays)
        else:
            raise ValueError("Incompatible observables!")

    @staticmethod
    def _map_entity_name(raw_name: str) -> str:
        if raw_name.startswith("Entity"):
            return raw_name[6:].lower()
        elif raw_name == "null":
            return raw_name
        else:
            print(f"Unknown entity prefix. Return raw name {raw_name}")
            return raw_name



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton

import os
import json

MC_ITEM_IDS = [
    "minecraft:acacia_boat",
    "minecraft:acacia_door",
    "minecraft:acacia_fence",
    "minecraft:acacia_fence_gate",
    "minecraft:acacia_stairs",
    "minecraft:activator_rail",
    "minecraft:air",
    "minecraft:anvil",
    "minecraft:apple",
    "minecraft:armor_stand",
    "minecraft:arrow",
    "minecraft:baked_potato",
    "minecraft:banner",
    "minecraft:barrier",
    "minecraft:beacon",
    "minecraft:bed",
    "minecraft:bedrock",
    "minecraft:beef",
    "minecraft:beetroot",
    "minecraft:beetroot_seeds",
    "minecraft:beetroot_soup",
    "minecraft:birch_boat",
    "minecraft:birch_door",
    "minecraft:birch_fence",
    "minecraft:birch_fence_gate",
    "minecraft:birch_stairs",
    "minecraft:black_glazed_terracotta",
    "minecraft:black_shulker_box",
    "minecraft:blaze_powder",
    "minecraft:blaze_rod",
    "minecraft:blue_glazed_terracotta",
    "minecraft:blue_shulker_box",
    "minecraft:boat",
    "minecraft:bone",
    "minecraft:bone_block",
    "minecraft:book",
    "minecraft:bookshelf",
    "minecraft:bow",
    "minecraft:bowl",
    "minecraft:bread",
    "minecraft:brewing_stand",
    "minecraft:brick",
    "minecraft:brick_block",
    "minecraft:brick_stairs",
    "minecraft:brown_glazed_terracotta",
    "minecraft:brown_mushroom",
    "minecraft:brown_mushroom_block",
    "minecraft:brown_shulker_box",
    "minecraft:bucket",
    "minecraft:cactus",
    "minecraft:cake",
    "minecraft:carpet",
    "minecraft:carrot",
    "minecraft:carrot_on_a_stick",
    "minecraft:cauldron",
    "minecraft:chain_command_block",
    "minecraft:chainmail_boots",
    "minecraft:chainmail_chestplate",
    "minecraft:chainmail_helmet",
    "minecraft:chainmail_leggings",
    "minecraft:chest",
    "minecraft:chest_minecart",
    "minecraft:chicken",
    "minecraft:chorus_flower",
    "minecraft:chorus_fruit",
    "minecraft:chorus_fruit_popped",
    "minecraft:chorus_plant",
    "minecraft:clay",
    "minecraft:clay_ball",
    "minecraft:clock",
    "minecraft:coal",
    "minecraft:coal_block",
    "minecraft:coal_ore",
    "minecraft:cobblestone",
    "minecraft:cobblestone_wall",
    "minecraft:command_block",
    "minecraft:command_block_minecart",
    "minecraft:comparator",
    "minecraft:compass",
    "minecraft:concrete",
    "minecraft:concrete_powder",
    "minecraft:cooked_beef",
    "minecraft:cooked_chicken",
    "minecraft:cooked_fish",
    "minecraft:cooked_mutton",
    "minecraft:cooked_porkchop",
    "minecraft:cooked_rabbit",
    "minecraft:cookie",
    "minecraft:crafting_table",
    "minecraft:cyan_glazed_terracotta",
    "minecraft:cyan_shulker_box",
    "minecraft:dark_oak_boat",
    "minecraft:dark_oak_door",
    "minecraft:dark_oak_fence",
    "minecraft:dark_oak_fence_gate",
    "minecraft:dark_oak_stairs",
    "minecraft:daylight_detector",
    "minecraft:deadbush",
    "minecraft:detector_rail",
    "minecraft:diamond",
    "minecraft:diamond_axe",
    "minecraft:diamond_block",
    "minecraft:diamond_boots",
    "minecraft:diamond_chestplate",
    "minecraft:diamond_helmet",
    "minecraft:diamond_hoe",
    "minecraft:diamond_horse_armor",
    "minecraft:diamond_leggings",
    "minecraft:diamond_ore",
    "minecraft:diamond_pickaxe",
    "minecraft:diamond_shovel",
    "minecraft:diamond_sword",
    "minecraft:dirt",
    "minecraft:dispenser",
    "minecraft:double_plant",
    "minecraft:dragon_breath",
    "minecraft:dragon_egg",
    "minecraft:dropper",
    "minecraft:dye",
    "minecraft:egg",
    "minecraft:elytra",
    "minecraft:emerald",
    "minecraft:emerald_block",
    "minecraft:emerald_ore",
    "minecraft:enchanted_book",
    "minecraft:enchanting_table",
    "minecraft:end_bricks",
    "minecraft:end_crystal",
    "minecraft:end_portal_frame",
    "minecraft:end_rod",
    "minecraft:end_stone",
    "minecraft:ender_chest",
    "minecraft:ender_eye",
    "minecraft:ender_pearl",
    "minecraft:experience_bottle",
    "minecraft:farmland",
    "minecraft:feather",
    "minecraft:fence",
    "minecraft:fence_gate",
    "minecraft:fermented_spider_eye",
    "minecraft:filled_map",
    "minecraft:fire_charge",
    "minecraft:firework_charge",
    "minecraft:fireworks",
    "minecraft:fish",
    "minecraft:fishing_rod",
    "minecraft:flint",
    "minecraft:flint_and_steel",
    "minecraft:flower_pot",
    "minecraft:furnace",
    "minecraft:furnace_minecart",
    "minecraft:ghast_tear",
    "minecraft:glass",
    "minecraft:glass_bottle",
    "minecraft:glass_pane",
    "minecraft:glowstone",
    "minecraft:glowstone_dust",
    "minecraft:gold_block",
    "minecraft:gold_ingot",
    "minecraft:gold_nugget",
    "minecraft:gold_ore",
    "minecraft:golden_apple",
    "minecraft:golden_axe",
    "minecraft:golden_boots",
    "minecraft:golden_carrot",
    "minecraft:golden_chestplate",
    "minecraft:golden_helmet",
    "minecraft:golden_hoe",
    "minecraft:golden_horse_armor",
    "minecraft:golden_leggings",
    "minecraft:golden_pickaxe",
    "minecraft:golden_rail",
    "minecraft:golden_shovel",
    "minecraft:golden_sword",
    "minecraft:grass",
    "minecraft:grass_path",
    "minecraft:gravel",
    "minecraft:gray_glazed_terracotta",
    "minecraft:gray_shulker_box",
    "minecraft:green_glazed_terracotta",
    "minecraft:green_shulker_box",
    "minecraft:gunpowder",
    "minecraft:hardened_clay",
    "minecraft:hay_block",
    "minecraft:heavy_weighted_pressure_plate",
    "minecraft:hopper",
    "minecraft:hopper_minecart",
    "minecraft:ice",
    "minecraft:iron_axe",
    "minecraft:iron_bars",
    "minecraft:iron_block",
    "minecraft:iron_boots",
    "minecraft:iron_chestplate",
    "minecraft:iron_door",
    "minecraft:iron_helmet",
    "minecraft:iron_hoe",
    "minecraft:iron_horse_armor",
    "minecraft:iron_ingot",
    "minecraft:iron_leggings",
    "minecraft:iron_nugget",
    "minecraft:iron_ore",
    "minecraft:iron_pickaxe",
    "minecraft:iron_shovel",
    "minecraft:iron_sword",
    "minecraft:iron_trapdoor",
    "minecraft:item_frame",
    "minecraft:jukebox",
    "minecraft:jungle_boat",
    "minecraft:jungle_door",
    "minecraft:jungle_fence",
    "minecraft:jungle_fence_gate",
    "minecraft:jungle_stairs",
    "minecraft:ladder",
    "minecraft:lapis_block",
    "minecraft:lapis_ore",
    "minecraft:lava_bucket",
    "minecraft:lead",
    "minecraft:leather",
    "minecraft:leather_boots",
    "minecraft:leather_chestplate",
    "minecraft:leather_helmet",
    "minecraft:leather_leggings",
    "minecraft:leaves",
    "minecraft:leaves2",
    "minecraft:lever",
    "minecraft:light_blue_glazed_terracotta",
    "minecraft:light_blue_shulker_box",
    "minecraft:light_weighted_pressure_plate",
    "minecraft:lime_glazed_terracotta",
    "minecraft:lime_shulker_box",
    "minecraft:lingering_potion",
    "minecraft:lit_pumpkin",
    "minecraft:log",
    "minecraft:log2",
    "minecraft:magenta_glazed_terracotta",
    "minecraft:magenta_shulker_box",
    "minecraft:magma",
    "minecraft:magma_cream",
    "minecraft:map",
    "minecraft:melon",
    "minecraft:melon_block",
    "minecraft:melon_seeds",
    "minecraft:milk_bucket",
    "minecraft:minecart",
    "minecraft:mob_spawner",
    "minecraft:monster_egg",
    "minecraft:mossy_cobblestone",
    "minecraft:mushroom_stew",
    "minecraft:mutton",
    "minecraft:mycelium",
    "minecraft:name_tag",
    "minecraft:nether_brick",
    "minecraft:nether_brick_fence",
    "minecraft:nether_brick_stairs",
    "minecraft:nether_star",
    "minecraft:nether_wart",
    "minecraft:nether_wart_block",
    "minecraft:netherbrick",
    "minecraft:netherrack",
    "minecraft:noteblock",
    "minecraft:oak_stairs",
    "minecraft:observer",
    "minecraft:obsidian",
    "minecraft:orange_glazed_terracotta",
    "minecraft:orange_shulker_box",
    "minecraft:packed_ice",
    "minecraft:painting",
    "minecraft:paper",
    "minecraft:pink_glazed_terracotta",
    "minecraft:pink_shulker_box",
    "minecraft:piston",
    "minecraft:planks",
    "minecraft:poisonous_potato",
    "minecraft:porkchop",
    "minecraft:potato",
    "minecraft:potion",
    "minecraft:prismarine",
    "minecraft:prismarine_crystals",
    "minecraft:prismarine_shard",
    "minecraft:pumpkin",
    "minecraft:pumpkin_pie",
    "minecraft:pumpkin_seeds",
    "minecraft:purple_glazed_terracotta",
    "minecraft:purple_shulker_box",
    "minecraft:purpur_block",
    "minecraft:purpur_pillar",
    "minecraft:purpur_slab",
    "minecraft:purpur_stairs",
    "minecraft:quartz",
    "minecraft:quartz_block",
    "minecraft:quartz_ore",
    "minecraft:quartz_stairs",
    "minecraft:rabbit",
    "minecraft:rabbit_foot",
    "minecraft:rabbit_hide",
    "minecraft:rabbit_stew",
    "minecraft:rail",
    "minecraft:record_11",
    "minecraft:record_13",
    "minecraft:record_blocks",
    "minecraft:record_cat",
    "minecraft:record_chirp",
    "minecraft:record_far",
    "minecraft:record_mall",
    "minecraft:record_mellohi",
    "minecraft:record_stal",
    "minecraft:record_strad",
    "minecraft:record_wait",
    "minecraft:record_ward",
    "minecraft:red_flower",
    "minecraft:red_glazed_terracotta",
    "minecraft:red_mushroom",
    "minecraft:red_mushroom_block",
    "minecraft:red_nether_brick",
    "minecraft:red_sandstone",
    "minecraft:red_sandstone_stairs",
    "minecraft:red_shulker_box",
    "minecraft:redstone",
    "minecraft:redstone_block",
    "minecraft:redstone_lamp",
    "minecraft:redstone_ore",
    "minecraft:redstone_torch",
    "minecraft:reeds",
    "minecraft:repeater",
    "minecraft:repeating_command_block",
    "minecraft:rotten_flesh",
    "minecraft:saddle",
    "minecraft:sand",
    "minecraft:sandstone",
    "minecraft:sandstone_stairs",
    "minecraft:sapling",
    "minecraft:sea_lantern",
    "minecraft:shears",
    "minecraft:shield",
    "minecraft:shulker_shell",
    "minecraft:sign",
    "minecraft:silver_glazed_terracotta",
    "minecraft:silver_shulker_box",
    "minecraft:skull",
    "minecraft:slime",
    "minecraft:slime_ball",
    "minecraft:snow",
    "minecraft:snow_layer",
    "minecraft:snowball",
    "minecraft:soul_sand",
    "minecraft:spawn_egg",
    "minecraft:speckled_melon",
    "minecraft:spectral_arrow",
    "minecraft:spider_eye",
    "minecraft:splash_potion",
    "minecraft:sponge",
    "minecraft:spruce_boat",
    "minecraft:spruce_door",
    "minecraft:spruce_fence",
    "minecraft:spruce_fence_gate",
    "minecraft:spruce_stairs",
    "minecraft:stained_glass",
    "minecraft:stained_glass_pane",
    "minecraft:stained_hardened_clay",
    "minecraft:stick",
    "minecraft:sticky_piston",
    "minecraft:stone",
    "minecraft:stone_axe",
    "minecraft:stone_brick_stairs",
    "minecraft:stone_button",
    "minecraft:stone_hoe",
    "minecraft:stone_pickaxe",
    "minecraft:stone_pressure_plate",
    "minecraft:stone_shovel",
    "minecraft:stone_slab",
    "minecraft:stone_slab2",
    "minecraft:stone_stairs",
    "minecraft:stone_sword",
    "minecraft:stonebrick",
    "minecraft:string",
    "minecraft:structure_block",
    "minecraft:structure_void",
    "minecraft:sugar",
    "minecraft:tallgrass",
    "minecraft:tipped_arrow",
    "minecraft:tnt",
    "minecraft:tnt_minecart",
    "minecraft:torch",
    "minecraft:totem_of_undying",
    "minecraft:trapdoor",
    "minecraft:trapped_chest",
    "minecraft:tripwire_hook",
    "minecraft:vine",
    "minecraft:water_bucket",
    "minecraft:waterlily",
    "minecraft:web",
    "minecraft:wheat",
    "minecraft:wheat_seeds",
    "minecraft:white_glazed_terracotta",
    "minecraft:white_shulker_box",
    "minecraft:wooden_axe",
    "minecraft:wooden_button",
    "minecraft:wooden_door",
    "minecraft:wooden_hoe",
    "minecraft:wooden_pickaxe",
    "minecraft:wooden_pressure_plate",
    "minecraft:wooden_shovel",
    "minecraft:wooden_slab",
    "minecraft:wooden_sword",
    "minecraft:wool",
    "minecraft:writable_book",
    "minecraft:written_book",
    "minecraft:yellow_flower",
    "minecraft:yellow_glazed_terracotta",
    "minecraft:yellow_shulker_box",
]

ALL_ACHIEVEMENTS = [
    "achievement.openInventory",
    "achievement.mineWood",
    "achievement.buildWorkBench",
    "achievement.buildPickaxe",
    "achievement.buildFurnace",
    "achievement.acquireIron",
    "achievement.buildHoe",
    "achievement.makeBread",
    "achievement.bakeCake",
    "achievement.buildBetterPickaxe",
    "achievement.cookFish",
    "achievement.onARail",
    "achievement.buildSword",
    "achievement.killEnemy",
    "achievement.killCow",
    "achievement.flyPig",
    "achievement.snipeSkeleton",
    "achievement.diamonds",
    "achievement.diamondsToYou",
    "achievement.portal",
    "achievement.ghast",
    "achievement.blazeRod",
    "achievement.potion",
    "achievement.theEnd",
    "achievement.theEnd2",
    "achievement.enchantments",
    "achievement.overkill",
    "achievement.bookcase",
    "achievement.breedCow",
    "achievement.spawnWither",
    "achievement.killWither",
    "achievement.fullBeacon",
    "achievement.exploreAllBiomes",
    "achievement.overpowered",
]

KEYMAP = {
    "17": "forward",
    "30": "left",
    "31": "back",
    "32": "right",
    "57": "jump",
    "18": "inventory",
    "16": "drop",
    "42": "sneak",
    "29": "sprint",
    "-100": "attack",  # BUTTON0 Left Click
    "-99": "use",  # BUTTON1 Right Click
    "-98": "pickItem",  # BUTTON2 Middle Click
    # '20': 'chat',  # This and following not currently in use
    # '33': 'swapHands',
    # '15': 'playerlist',  # Show player list gui
    # '53': 'command',  # Start typing server cmd
    # '60': 'screenshot',
    # '63': 'togglePerspective',
    # '87': 'fullscreen',
    # '46': 'saveToolbarActivator',
    # '45': 'loadToolbarActivator',
    # '38': 'advancements',
}

KEYMAP.update({str(x + 1): str(x) for x in range(1, 10)})

# TODO: add all other keys.
INVERSE_KEYMAP = {KEYMAP[key]: key for key in KEYMAP}

MAX_LIFE = 20  # Actual max life can be greater (e.g. after eating a golden apple)
MAX_ARMOR = 20  # actual max armor value (i.e., wearing a full suit of diamond armors)
MAX_XP = 1395  # Represents level 30 in game - bounded by signed 32 bit int max
MAX_BREATH = 300  # Should be max encountered - subject to change by mods
MAX_FOOD = 20  # Default max food
MAX_FOOD_SATURATION = 20.0  # Current max saturation limit as of 1.11.2
MAX_SCORE = 0x7FFFFF  # Implemented as XP in survival but can change e.g. mini-games


def get_item_id(item: str) -> int:
    """
    Gets the item ID of an MC item.
    :param item: The item string
    :return: The internal ID of the item.
    """
    if not item.startswith("minecraft:"):
        item = "minecraft:" + item

    return MC_ITEM_IDS.index(item)


def get_key_from_id(id: str) -> str:
    """
    Gets the key from an id.
    :param id:
    :return:
    """
    assert id in KEYMAP, "ID not found"
    return KEYMAP[id]


mc_constants_file = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "mc_constants.json"
)
all_data = json.load(open(mc_constants_file))

ALL_ITEMS = sorted([item["type"] for item in all_data["items"]])
ALL_ITEM_VARIANTS = sorted(
    [
        f"{item['type']}_{variant}"
        for item in all_data["items"]
        if "variant" in item
        for variant in item["variant"]
    ]
)

ALL_STATS = sorted([stat["statID"] for stat in all_data["stats"]])
ALL_STAT_KEYS = sorted([stat["minerl_keys"] for stat in all_data["stats"]])

# We choose these not to be included by default; they are not items.
NONE = "none"
INVALID = "invalid"

# TODO remove hack based on ordering in MineRL
MINERL_ITEM_MAP = sorted(["none"] + ALL_ITEMS)

ITEMS_BY_CATEGORY = {
    # Items which take 2 seconds to USE
    "edible": [
        item["type"]
        for item in all_data["items"]
        if item["useAction"] in {"EAT", "DRINK"}
    ],
    # Items which have ongoing effect when equipped (includes armor)
    "tool": [
        item["type"] for item in all_data["items"] if item["tab"] in {"tools", "combat"}
    ],
    # Items which are used for building
    "building_block": [
        item["type"] for item in all_data["items"] if item["tab"] in {"buildingBlock"}
    ],
    # Redstone items (doors, buttons, levers)
    "redstone": [
        item["type"] for item in all_data["items"] if item["tab"] in {"redstone"}
    ],
    # Brewing items
    "brewing": [
        item["type"] for item in all_data["items"] if item["tab"] in {"brewing"}
    ],
    # Decoration items (includes torch)
    "decoration": [
        item["type"] for item in all_data["items"] if item["tab"] in {"decoration"}
    ],
}


PLACEABLE_ITEM_NAMES = sorted(
    [
        item["type"].replace("_", " ")
        for item in all_data["items"]
        if item["block"] is True
    ]
)


# Check that all edible items have the same maxUseDuration
use_times = {
    item["maxUseDuration"]
    for item in all_data["items"]
    if item["useAction"] in {"EAT", "DRINK"}
}
assert len(use_times) == 1, "Edible items with multiple different eating times."
EDIBLE_USE_TICKS = use_times.pop()


def recursive_dict_eq(d1, d2):
    if isinstance(d1, dict) != isinstance(d2, dict):
        return False
    if isinstance(d1, dict):
        if set(d1.keys()) != set(d2.keys()):
            return False
        return all([recursive_dict_eq(d1[k], d2[k]) for k in d1.keys()])
    else:
        return d1 == d2


def duplicate_dict_in_list(dict, list):
    for item in list:
        if recursive_dict_eq(dict, item):
            return True
    return False


def dedup_list(dicts):
    """
    Takes a list of dictionary objects and removes any duplicates (compared recursively by value).
    """
    result = []
    for next_dict in dicts:
        if not duplicate_dict_in_list(next_dict, result):
            result.append(next_dict)
    return result


def sort_recipes_by_output(json):
    result = {item: [] for item in ALL_ITEMS}
    for recipe in json:
        if len(recipe["ingredients"]) == 0:
            # Empty recipe
            continue
        if recipe["outputItemName"] in recipe["ingredients"]:
            # Circular recipe
            continue
        result[recipe["outputItemName"]].append(recipe)
    for item, recipes in result.items():
        result[item] = dedup_list(recipes)
    return result


CRAFTING_RECIPES_BY_OUTPUT = sort_recipes_by_output(all_data["craftingRecipes"])
SMELTING_RECIPES_BY_OUTPUT = sort_recipes_by_output(all_data["smeltingRecipes"])

ALL_PERSONAL_CRAFTING_ITEMS = sorted(
    [
        item["type"]
        for item in all_data["items"]
        if item["type"] in CRAFTING_RECIPES_BY_OUTPUT.keys()
        and len(CRAFTING_RECIPES_BY_OUTPUT[item["type"]]) > 0
        and all(
            [
                recipe["recipeSize"] in [0, 1, 2, 4]  # TODO recipeSize needs to be 2D
                for recipe in CRAFTING_RECIPES_BY_OUTPUT[item["type"]]
            ]
        )
    ]
)

ALL_CRAFTING_TABLE_ITEMS = sorted(
    [
        item["type"]
        for item in all_data["items"]
        if item["type"] in CRAFTING_RECIPES_BY_OUTPUT.keys()
        and len(CRAFTING_RECIPES_BY_OUTPUT[item["type"]]) > 0
        and any(
            [
                recipe["recipeSize"] <= 9
                for recipe in CRAFTING_RECIPES_BY_OUTPUT[item["type"]]
            ]
        )
    ]
)

ALL_CRAFTING_TABLE_ONLY_ITEMS = sorted(
    [x for x in ALL_CRAFTING_TABLE_ITEMS if x not in ALL_PERSONAL_CRAFTING_ITEMS]
)

ALL_SMELTING_ITEMS = sorted(
    [
        item["type"]
        for item in all_data["items"]
        if item["type"] in SMELTING_RECIPES_BY_OUTPUT.keys()
        and len(SMELTING_RECIPES_BY_OUTPUT[item["type"]]) > 0
    ]
)

# all self-craftable/table craftable/smeltable items, including "none"
ALL_HAND_CRAFT_ITEMS_NN_ACTIONS = [
    x
    for x in ALL_PERSONAL_CRAFTING_ITEMS
    if x
    not in [
        "coal",
        "diamond",
        "emerald",
        "gold_ingot",
        "iron_ingot",
        "redstone",
        "stone",
    ]
]
ALL_TABLE_CRAFT_ONLY_ITEMS_NN_ACTIONS = [
    x for x in ALL_CRAFTING_TABLE_ONLY_ITEMS if x not in ["iron_ingot", "gold_ingot"]
]
ALL_SMELT_ITEMS_NN_ACTIONS = [
    x for x in ALL_SMELTING_ITEMS if x not in ["iron_nugget", "dye", "gold_nugget"]
]
ALL_CRAFT_SMELT_ITEMS = (
    ALL_HAND_CRAFT_ITEMS_NN_ACTIONS
    + ALL_TABLE_CRAFT_ONLY_ITEMS_NN_ACTIONS
    + ALL_SMELT_ITEMS_NN_ACTIONS
)

assert len(set(ALL_CRAFT_SMELT_ITEMS)) == len(ALL_HAND_CRAFT_ITEMS_NN_ACTIONS) + len(
    ALL_TABLE_CRAFT_ONLY_ITEMS_NN_ACTIONS
) + len(ALL_SMELT_ITEMS_NN_ACTIONS)

EQUIPMENT_SLOTS = [
    "mainhand",
    "offhand",
    "feet",
    "legs",
    "chest",
    "head",
]

BEST_ITEMS_PER_EQUIPMENT_SLOT = {
    equip: [
        item["type"] for item in all_data["items"] if item["bestEquipmentSlot"] == equip
    ]
    for equip in EQUIPMENT_SLOTS
}

NON_MAINHAND_ITEM_NAMES = []
for k, v in BEST_ITEMS_PER_EQUIPMENT_SLOT.items():
    if k != "mainhand":
        NON_MAINHAND_ITEM_NAMES.extend([item.replace("_", " ") for item in v])

N_INV_SLOTS = 36  # including main-hand

MS_PER_STEP = 50
STEPS_PER_MS = 1 / 50


def strip_item_prefix(minecraft_name):
    # Names in minecraft start with 'minecraft:', like:
    # 'minecraft:log', or 'minecraft:cobblestone'
    if minecraft_name.startswith("minecraft:"):
        return minecraft_name[len("minecraft:") :]

    return minecraft_name


# Transcribed from: https://minecraft.gamepedia.com/index.php?title=Biome&oldid=1057817,
#   which has the biomes as they were during the minecraft version at the time: 1.11.2
# _m indicates that the terrain is more mountainous
# _f indicates that the terrain has trees where the base biome does not.
BIOMES_MAP = {
    "ice_plains": 12,
    "ice_plains_spikes": 140,
    "cold_taiga": 30,
    "cold_taiga_m": 158,
    "frozen_river": 11,
    "cold_beach": 26,
    "extreme_hills": 3,
    "extreme_hills_m": 131,
    "extreme_hills_plus": 34,
    "extreme_hills_plus_m": 162,
    "taiga": 5,
    "taiga_m": 133,
    "mega_taiga": 32,
    "mega_spruce_taiga": 160,
    "stone_beach": 25,
    "plains": 1,
    "sunflower_plains": 129,
    "forest": 4,
    "flower_forest": 132,
    "birch_forest": 27,
    "birch_forest_m": 155,
    "roofed_forest": 29,
    "roofed_forest_m": 157,
    "swampland": 6,
    "swampland_m": 134,
    "jungle": 21,
    "jungle_m": 149,
    "jungle_edge": 23,
    "jungle_edge_m": 151,
    "river": 7,
    "beach": 16,
    "mushroom_island": 14,
    "mushroom_island_shore": 15,
    "desert": 2,
    "desert_m": 130,
    "savanna": 35,
    "savanna_m": 163,
    "mesa": 37,
    "mesa_bryce": 165,
    "mesa_plateau_f": 38,
    "mesa_plateau_f_m": 166,
    "savanna_plateau": 36,
    "mesa_plateau": 39,
    "savanna_plateau_m": 164,
    "mesa_plateau_m": 167,
    "ocean": 0,
    "deep_ocean": 24,
    "ice_mountains": 13,
    "desert_hills": 17,
    "forest_hills": 18,
    "taiga_hills": 19,
    "jungle_hills": 22,
    "birch_forest_hills": 28,
    "cold_taiga_hills": 31,
    "mega_taiga_hills": 33,
    "tall_birch_hills": 156,
    "mega_spruce_taiga_hills": 161,
    "plains_m": 128,
    "frozen_ocean": 10,
    "extreme_hills_edge": 20,
    "the_end": 9,
    "nether": 8,
    "the_void": 127,
}



import numpy as np

from ...sim.mc_meta import mc as MC


__all__ = ["get_recipes_matrix", "get_inventory_vector"]


def get_recipes_matrix():
    recipes = []
    for item in MC.ALL_CRAFT_SMELT_ITEMS:
        ingredients = (
            MC.CRAFTING_RECIPES_BY_OUTPUT[item]
            if item
            in MC.ALL_HAND_CRAFT_ITEMS_NN_ACTIONS
            + MC.ALL_TABLE_CRAFT_ONLY_ITEMS_NN_ACTIONS
            else MC.SMELTING_RECIPES_BY_OUTPUT[item]
        )
        # take the first recipe if multiple recipes
        ingredients = ingredients[0]["ingredients"]
        recipe_vector = np.zeros(len(MC.ALL_ITEMS))
        for k, v in ingredients.items():
            recipe_vector[MC.ALL_ITEMS.index(k)] += v
        recipes.append(recipe_vector)
    recipes = np.stack(recipes, axis=0)
    assert recipes.shape == (
        len(MC.ALL_CRAFT_SMELT_ITEMS),
        len(MC.ALL_ITEMS),
    ), "Internal"
    return recipes


def get_inventory_vector(inventory):
    vec = np.zeros(len(MC.ALL_ITEMS))
    items, quantities = inventory["name"], inventory["quantity"]
    for item, quantity in zip(items, quantities):
        if " " in item:
            item = item.replace(" ", "_")
        vec[MC.ALL_ITEMS.index(item)] += quantity
    return vec[np.newaxis, ...]



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
import logging

from minedojo.sim.handlers.agent import action
from minedojo.sim.handlers import util
import minedojo.sim.mc_meta.mc as mc


class EquipAction(action.ItemWithMetadataListAction):
    """
    An action handler for observing a list of equipped items
    """

    logger = logging.getLogger(__name__ + ".EquipAction")

    def __init__(self, items: list, _default="none", _other="other"):
        """
        Initializes the space of the handler to be one for each item in the list plus one for the
        default no-craft action
        """
        super().__init__("equip", items, _default=_default, _other=_other)
        self._previous_id = self._default
        self._previous_type = None
        self._previous_metadata = None

    def xml_template(self) -> str:
        return str("<EquipCommands/>")

    def reset(self):
        self._previous_id = self._default

    def from_universal(self, obs) -> str:
        slots_gui_type = obs["slots"]["gui"]["type"]
        if slots_gui_type == "class net.minecraft.inventory.ContainerPlayer":
            hotbar_index = int(obs["hotbar"])
            hotbar_slot = obs["slots"]["gui"]["slots"][-10 + hotbar_index]

            if len(hotbar_slot.keys()) == 0:
                # Really we should be returning air, and requiring that it exists in `self.items`?
                return self._default

            item_type = mc.strip_item_prefix(hotbar_slot["name"])
            metadata = hotbar_slot["variant"]
            assert metadata in range(16)
            id = util.get_unique_matching_item_list_id(
                self.items, item_type, metadata, clobber_logs=False
            )
            if id is None:
                # Matching item ID not found because item_type + metadata is outside of item list.
                # Return "other" if this combination
                # of item_type + metadata is different than the previous item step. (New other item
                # equipped.)
                # Otherwise, return "none" to indicate that the equipped item stack has not changed.
                if (
                    self._previous_type == item_type
                    and self._previous_metadata == metadata
                ):
                    result = self._default
                else:
                    result = self._other
            elif id == self._previous_id:
                result = self._default
            else:
                result = id

            self._previous_id = id
            self._previous_type = item_type
            self._previous_metadata = metadata
            return result
        else:
            expected_ignore_types = (  # Filter these out to reduce stderr clutter
                "class net.minecraft.inventory.ContainerWorkbench",
                "class net.minecraft.inventory.ContainerFurnace",
                "class net.minecraft.inventory.ContainerChest",
                "class net.minecraft.inventory.ContainerMerchant",
            )

            if slots_gui_type not in expected_ignore_types:
                self.logger.debug(
                    f"Unexpected slots_gui_type={slots_gui_type}, "
                    f"Abandoning processing and simply returning '{self._default}'"
                )
            return self._default



from __future__ import annotations

import json
import os

import praw

from .download import download as dl
from .download import get_fn


class RedditDataset:
    """
    Class for MineDojo Reddit Database API.
    We follow PyTorch Dataset format but without actually inheriting from PyTorch dataset to keep the framework general.
    See https://praw.readthedocs.io/en/stable/getting_started/quick_start.html
    for setting up ``client_id``, ``cliend_secret`` and ``user_agent``.

    Args:
        download: If ``True`` and there is no existing cache directory, the data will be downloaded automatically.

        download_dir: Directory path where the downloaded data will be saved.
                Default: ``~/.minedojo/``.

        client_id: The client ID to access Reddit’s API as a script application.

        client_secret: The client secret to access Reddit’s API as a script application.

        user_agent:  A unique identifier that helps Reddit determine the source of network requests.

        max_comments: Maximum number of comments to load.

    Examples:
        >>> from minedojo.data import RedditDataset
        >>> reddit_dataset = RedditDataset(client_id={your_client_id}, client_secret={your_client_secret}, user_agent={your_user_agent})
        >>> print(reddit_dataset[0].keys())
        dict_keys(['id', 'title', 'link', 'score', 'num_comments', 'created_utc', 'type', 'content', 'comments'])
    """

    def __init__(
        self,
        *,
        download: bool = True,
        download_dir: None | str = None,
        client_id: str = None,
        client_secret: str = None,
        user_agent: str = None,
        max_comments: int = 100,
    ):
        if download_dir is None:
            download_dir = os.path.join(os.path.expanduser("~"), ".minedojo")

        if download:
            self.root = dl("reddit", download_dir)
        else:
            self.root, _, url = get_fn("reddit", download_dir)
            assert os.path.exists(self.root), (
                f"Reddit data file {self.root} does not exist. "
                "Please set download=True or you can manually "
                f"download it from {url}."
            )

        with open(self.root, "r") as f:
            self.data = json.load(f)

        if client_id is None or client_secret is None or user_agent is None:
            raise RedditAPIKeyNotSpecifiedError

        self.api = praw.Reddit(
            client_id=client_id,
            client_secret=client_secret,
            user_agent=user_agent,
            check_for_async=False,
        )

        self.max_comments = max_comments

    def get_metadata(self, post_id: str, post_type: str) -> dict:
        """Get post metadata using PRAW.

        Args:
            post_id: The unique, base36 ID of a Reddit post.
            post_type: The type of the post, either "image", "text", "video" or "link".

        Return:
            A dictionary containing the metadata of the post.

            - id(``str``) - The unique, base36 Reddit post ID.

            - title(``str``) - The title of the Reddit post.

            - link(``str``) - The url of the Reddit post.

            - score(``int``) - The score of the Reddit post.

            - num_comments(``int``) - The number of comments under the Reddit post. Does not account for deleted comments.

            - created_utc(``int``) - The date and time the Reddit post was created, in UTC format.

            - type(``str``) - The type of the post, either "image", "text", "video" or "link".

            - content(``str``) - If text type post, text in post body. Otherwise, the media source url or website link.

            - comments(``list[dict]``)

                - id(``str``) - The unique base36 comment ID.

                - parent_id(``str``) - The ID of the comment's parent in the nested comment tree.

                - content(``str``) - The text in comment body.
        """
        post = self.api.submission(id=post_id)

        metadata = {
            "id": post.id,
            "title": post.title,
            "link": f"https://www.reddit.com/r/Minecraft/comments/{post.id}",
            "score": post.score,
            "num_comments": post.num_comments,
            "created_utc": post.created_utc,
            "type": post_type,
            "content": post.selftext if post_type == "text" else post.url,
            "comments": self.get_comments(post),
        }

        return metadata

    def get_comments(self, post: praw.models.Submission) -> list[dict]:
        comments = []
        comment_queue = post.comments[:]
        while len(comments) < self.max_comments and comment_queue:
            comment = comment_queue.pop(0)
            if isinstance(comment, praw.models.MoreComments):
                comment_queue.extend(comment.comments())
                continue

            comments.append(
                {
                    "id": comment.id,
                    "parent_id": comment.parent_id[3:],
                    "content": comment.body,
                }
            )

        return comments

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.get_metadata(
            post_id=self.data[idx]["id"], post_type=self.data[idx]["type"]
        )


class RedditAPIKeyNotSpecifiedError(Exception):
    def __init__(self):
        self.message = (
            'You need to specify "client_id", "client_secret" and "user_agent" for Reddit API. '
            "You can refer to https://praw.readthedocs.io/en/stable/getting_started/quick_start.html "
            "for the instructions of obtaining Reddit API keys."
        )
        super().__init__(self.message)



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
from minedojo.sim.handlers.agent.actions.craft import CraftAction


class SmeltAction(CraftAction):
    """
    An action handler for crafting items when agent is in view of a crafting table

    Note when used along side Craft Item, block lists must be disjoint or from_universal will fire multiple times
    """

    _command = "smeltNearby"

    def to_string(self):
        return "smelt"

    def xml_template(self) -> str:
        return str("<NearbySmeltCommands/>")

    def from_universal(self, obs):
        if (
            "diff" in obs
            and "smelted" in obs["diff"]
            and len(obs["diff"]["smelted"]) > 0
        ):
            try:
                x = self._univ_items.index(obs["diff"]["smelted"][0]["item"])
                return obs["diff"]["smelted"][0]["item"].split("minecraft:")[-1]
            except ValueError:
                return self._default
                # return self._items.index('other')
        else:
            return self._default



import coloredlogs
import argparse
import time
import shutil
import logging
import psutil
import subprocess
import os
import sys
from daemoniker import daemonize

logger = logging.getLogger("watchdog")
WATCHDOG_DIR = os.path.join("logs", "watchdog")

CHILD_DIR_ARG = "child-dirs"


def parse_args():
    """Parses the arguments for the process watcher."""
    parser = argparse.ArgumentParser(
        description="A general process watcher utility that ensures "
        "that a parent and child process terminate uniformly."
    )
    parser.add_argument("parent_pid", type=int, help="The PID of the parent process.")
    parser.add_argument("child_pid", type=int, help="The PID of the child process.")
    parser.add_argument(
        f"--{CHILD_DIR_ARG}",
        type=str,
        help="Temporary directories which should be deleted "
        "were the processes to terminate.",
        nargs="+",
    )
    return parser.parse_args()


def launch(parent_pid, child_pid, *temp_dirs):
    """Launches the process watcher with a PID.

    Args:
        parent_pid (int): The parent PID.
        child_pid (int): The child PID.

    Returns:
        psutil.Process: The process object of the watcher.
    """
    logger.info("Launching watchdog daemonizer.")
    subprocess.check_call(
        [
            sys.executable,
            "-m",
            "minedojo.sim.bridge.utils.watchdog",
            str(parent_pid),
            str(child_pid),
            f"--{CHILD_DIR_ARG}",
        ]
        + list(temp_dirs)
    )
    logger.info("Watchdog launched successfully.")


def reap_process_and_children(process, timeout=5):
    "Tries hard to terminate and ultimately kill all the children of this process."

    def on_process_wait_successful(proc):
        returncode = proc.returncode if hasattr(proc, "returncode") else None
        logger.info("Process {} terminated with exit code {}".format(proc, returncode))

    def get_process_info(proc):
        return f"{proc.pid}:{proc.name()}:{proc.exe()} i {proc.status()}, owner {proc.ppid()}"

    procs = process.children(recursive=True)[::-1] + [process]
    try:
        logger.info(
            f"About to reap process tree of {get_process_info(process)}, printing process tree status in termination order:"
        )
        for p in procs:
            logger.info(f"\t-{get_process_info(p)}")
    except psutil.ZombieProcess:
        logger.info("Zombie process found in process tree.")

    for p in procs:
        try:
            logger.info(f"Trying to SIGTERM {get_process_info(p)}")
            p.terminate()
            try:
                p.wait(timeout=timeout)
                on_process_wait_successful(p)
            except psutil.TimeoutExpired:
                logger.info(
                    f"Process {p.pid} survived SIGTERM; trying SIGKILL on {get_process_info(p)}"
                )
                p.kill()

                try:
                    p.wait(timeout=timeout)
                    on_process_wait_successful(p)
                except psutil.TimeoutExpired:
                    # Give up
                    logger.info(
                        f"Process {p.pid} survived SIGKILL; giving up (final status) {p.status()}, (owner) {p.ppid()}"
                    )
        except (psutil.NoSuchProcess, psutil.ZombieProcess):
            logger.info(f"Process {p} does not exist or is zombie.")


def main(args):
    """The primary watchdog.

    Args:
        args (--): Arguments from the process watcher.
    """

    # Wait for processes to be launched
    logger.info(
        f"Watchdog started between parent {args.parent_pid} and child {args.child_pid}"
    )
    time.sleep(1)
    try:
        child = psutil.Process(args.child_pid)
    except psutil.NoSuchProcess:
        child = None
    try:
        parent = psutil.Process(args.parent_pid)
    except psutil.NoSuchProcess:
        parent = None

    while True:
        try:
            # Sleep for a short time, and check if subprocesses needed to be killed.
            time.sleep(0.5)

            if not parent.is_running() or parent is None:
                logger.info("Parent is not running, hence we need terminate the child.")
                if child is not None:
                    reap_process_and_children(child)
                    try:
                        for temp_dir in args[CHILD_DIR_ARG]:
                            shutil.rmtree(temp_dir)
                    except OSError:
                        logger.warning(
                            "Failed to delete temporary child directory. It may have already been removed."
                        )
                    except KeyError:
                        # No CHILD_DIR_ARG provided.
                        pass
                return
            # Kill the watcher if the child is no longer running.
            # If you want to attempt to restart the child on failure, this
            # would be the location to do so.
            if not child.is_running():
                logger.info("Child is not running anymore, launcher can terminate.")
                return
        except KeyboardInterrupt:
            pass


if __name__ == "__main__":
    args = parse_args()
    os.makedirs(WATCHDOG_DIR, exist_ok=True)

    os_cur_dir = os.path.abspath(os.getcwd())
    watcher_name = f"watchdog{args.parent_pid}-{args.child_pid}"

    daemonize(os.path.join(os_cur_dir, WATCHDOG_DIR, watcher_name + ".pid"))

    coloredlogs.install(
        level=logging.DEBUG,
        stream=open(os.path.join(os_cur_dir, WATCHDOG_DIR, watcher_name + ".log"), "w"),
    )

    main(args)
    exit()  # Correctly invoke the at_exit handlers and any other os clean up _explicitly_!



from typing import List, Optional, Union, Tuple, Dict

from .base import ExtraSpawnMetaTaskBase
from ...sim.inventory import InventoryItem
from .extra_spawn import SpawnItem2Condition
from .utils import (
    always_satisfy_condition,
    simple_stat_kill_entity_based_check,
    simple_stat_kill_entity_based_reward,
)


class CombatMeta(ExtraSpawnMetaTaskBase):
    """
    Class for combat tasks.
    Args:
        allow_mob_spawn: If ``True``, allow mobs (animals and hostiles) to spawn.
                Default: ``True``.

        always_night: If ``True``, the world will always be at night.
                Default: ``False``.

        break_speed_multiplier: Controls the speed of breaking blocks. A value larger than 1.0 accelerates the breaking.
                Default: ``1.0``.

        drawing_str: Draws shapes (e.g. spheres, cuboids) in the minecraft world.
                Default: ``None``.

        event_level_control: If ``True``, the agent is able to perform high-level controls including place and equip.
                If ``False``, then is keyboard-mouse level control.
                Default: ``True``.

        fast_reset: If ``True``, reset using MC native command `/kill`, instead of waiting for re-generating new worlds.
            Default: ``True``.

            .. warning::
                Side effects:

                1. Changes to the world will not be reset. E.g., if the agent chops lots of trees then calling
                fast reset will not restore those trees.

                2. If you specify agent starting health and food, these specs will only be respected at the first reset
                (i.e., generating a new world) but will not be respected in the following resets (i.e., reset using MC cmds).
                So be careful to use this wrapper if your usages require specific initial health/food.

                3. Statistics/achievements will not be reset. This wrapper will maintain a property ``info_prev_reset``.
                If your tasks use stat/achievements to evaluation, please retrieve this property and compute differences.

        image_size: The size of image observations.

        initial_inventory: If not ``None``, specifies initial items in the agent's inventory.
                Use class ``InventoryItem`` to specify items.
                Default: ``None``.

        initial_mobs: The types of mobs that are spawned initially.
                Default: ``None``.

        initial_mob_spawn_range_high: The upper bound on each horizontal axis from the center of the area to spawn initially.
                Default: ``None``.

        initial_mob_spawn_range_low: The lower bound on each horizontal axis from the center of the area to spawn initially.
                Default: ``None``.

        initial_weather: If not ``None``, specifies the initial weather.
                Can be one of ``"clear"``, ``"normal"``, ``"rain"``, ``"thunder"``.
                Default: ``None``.

        lidar_rays: Defines the directions and maximum distances of the lidar rays if ``use_lidar`` is ``True``.
                If supplied, should be a list of tuple(pitch, yaw, distance).
                Pitch and yaw are in radians and relative to agent looking vector.
                Default: ``None``.

        reward_weights: The reward weight for each target in the task.
                Default: ``1.0``.

        seed: The seed for an instance's internal generator.
                Default: ``None``.

        sim_name: Name of a simulation instance.
                Default: ``"CombatMeta"``.

        spawn_range_high: The upper bound on each horizontal axis from the center of the area to spawn
                Default: ``None``.

        spawn_range_low: The lower bound on each horizontal axis from the center of the area to spawn
                Default: ``None``.

        spawn_rate: The probability of spawning in each step
                Default: ``None``.

        specified_biome: The specified biome of the task.
                Default: ``None``.

        start_at_night: If ``True``, the task starts at night.
                Default: ``True``.

        start_food: If not ``None``, specifies initial food condition of the agent.
                Default: ``None``.

        start_health: If not ``None``, specifies initial health condition of the agent.
                Default: ``None``.

        start_position: If not ``None``, specifies the agent's initial location and orientation in the minecraft world.
                Default: ``None``.

        use_lidar: If ``True``, includes lidar in observations.
                Default: ``False``.

        use_voxel: If ``True``, includes voxel in observations.
                Default: ``False``.

        voxel_size: Defines the voxel's range in each axis if ``use_voxel`` is ``True``.
                If supplied, should be a dict with keys ``xmin``, ``xmax``, ``ymin``, ``ymax``, ``zmin``, ``zmax``.
                Each value specifies the voxel size relative to the agent.
                Default: ``None``.

        world_seed: Seed for deterministic world generation
                if ``generate_world_type`` is ``"default"`` or ``"specified_biome"``.
                See `here <https://minecraft.fandom.com/wiki/Seed_(level_generation)>`_ for more details.
                Default: ``None``.
    """

    _prompt_template = "Hunt {targets}."

    def __init__(
        self,
        *,
        # ------ hunting targets, quantities, and reward weights ------
        target_names: Union[str, List[str]],
        target_quantities: Union[int, List[int], Dict[str, int]],
        reward_weights: Union[int, float, Dict[str, Union[int, float]]] = 1.0,
        # ------ initial & extra spawn control ------
        initial_mobs: Optional[Union[str, List[str]]] = None,
        initial_mob_spawn_range_low: Optional[Tuple[int, int, int]] = None,
        initial_mob_spawn_range_high: Optional[Tuple[int, int, int]] = None,
        spawn_rate: Optional[
            Union[float, int, List[Union[float, int]], Dict[str, Union[float, int]]]
        ] = None,
        spawn_range_low: Optional[Tuple[int, int, int]] = None,
        spawn_range_high: Optional[Tuple[int, int, int]] = None,
        # ------ initial conditions ------
        start_position: Optional[Dict[str, Union[float, int]]] = None,
        start_health: Optional[float] = None,
        start_food: Optional[int] = None,
        start_at_night: bool = True,
        initial_weather: Optional[str] = None,
        initial_inventory: Optional[List[InventoryItem]] = None,
        drawing_str: Optional[str] = None,
        # ------ global conditions ------
        specified_biome: Optional[Union[int, str]] = None,
        always_night: bool = False,
        allow_mob_spawn: bool = True,
        break_speed_multiplier: float = 1.0,
        # ------ sim seed & world seed ------
        seed: Optional[int] = None,
        world_seed: Optional[str] = None,
        # ------ reset mode ------
        fast_reset: bool = True,
        fast_reset_random_teleport_range: Optional[int] = None,
        # ------ obs ------
        image_size: Union[int, Tuple[int, int]],
        use_voxel: bool = False,
        voxel_size: Optional[Dict[str, int]] = None,
        use_lidar: bool = False,
        lidar_rays: Optional[List[Tuple[float, float, float]]] = None,
        # ------ event-level action or keyboard-mouse level action ------
        event_level_control: bool = True,
        # ------ misc ------
        sim_name: str = "CombatMeta",
    ):
        if isinstance(target_names, str):
            target_names = [target_names]
        if isinstance(target_quantities, int):
            target_quantities = {k: target_quantities for k in target_names}
        elif isinstance(target_quantities, list):
            assert len(target_names) == len(target_quantities)
            target_quantities = {
                k: target_quantities[i] for i, k in enumerate(target_names)
            }
        elif isinstance(target_quantities, dict):
            assert set(target_names) == set(target_quantities.keys())
        if isinstance(reward_weights, int) or isinstance(reward_weights, float):
            reward_weights = {k: reward_weights for k in target_names}
        elif isinstance(reward_weights, dict):
            assert set(target_names) == set(reward_weights.keys())
        self._target_quantities = target_quantities

        spawn_condition = None
        if spawn_rate is not None:
            if isinstance(spawn_rate, float) or isinstance(spawn_rate, int):
                spawn_rate = {k: spawn_rate for k in target_names}
            elif isinstance(spawn_rate, list):
                assert len(spawn_rate) == len(target_names)
                spawn_rate = {k: spawn_rate[i] for i, k in enumerate(target_names)}
            elif isinstance(spawn_rate, dict):
                # don't do any checks here, so users can specify arbitrary spawn rates
                pass
            spawn_condition = {
                k: SpawnItem2Condition.get(k, always_satisfy_condition)
                for k in spawn_rate.keys()
            }

        success_criteria = [
            simple_stat_kill_entity_based_check(name=k, quantity=v)
            for k, v in target_quantities.items()
        ]
        reward_fns = [
            simple_stat_kill_entity_based_reward(name=k, weight=v)
            for k, v in reward_weights.items()
        ]

        start_time, allow_time_passage = None, True
        if start_at_night:
            start_time = 13000
        if always_night:
            start_time = 18000
            allow_time_passage = False

        generate_world_type = (
            "default" if specified_biome is None else "specified_biome"
        )

        super().__init__(
            initial_mobs=initial_mobs,
            initial_mob_spawn_range_low=initial_mob_spawn_range_low,
            initial_mob_spawn_range_high=initial_mob_spawn_range_high,
            extra_spawn_rate=spawn_rate,
            extra_spawn_condition=spawn_condition,
            extra_spawn_range_low=spawn_range_low,
            extra_spawn_range_high=spawn_range_high,
            fast_reset=fast_reset,
            fast_reset_random_teleport_range=fast_reset_random_teleport_range,
            success_criteria=success_criteria,
            reward_fns=reward_fns,
            seed=seed,
            sim_name=sim_name,
            image_size=image_size,
            use_voxel=use_voxel,
            voxel_size=voxel_size,
            use_lidar=use_lidar,
            lidar_rays=lidar_rays,
            event_level_control=event_level_control,
            initial_inventory=initial_inventory,
            break_speed_multiplier=break_speed_multiplier,
            world_seed=world_seed,
            start_position=start_position,
            initial_weather=initial_weather,
            start_time=start_time,
            allow_time_passage=allow_time_passage,
            start_health=start_health,
            start_food=start_food,
            drawing_str=drawing_str,
            allow_mob_spawn=allow_mob_spawn,
            generate_world_type=generate_world_type,
            specified_biome=specified_biome,
        )

    @property
    def task_prompt(self) -> str:
        filling = ", ".join(
            [
                f"{v} {str(k).replace('_', ' ')}"
                for k, v in self._target_quantities.items()
            ]
        )
        return super().get_prompt(targets=filling)



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
from typing import Optional

from minedojo.sim.handlers.agent.action import ItemListAction


class CraftAction(ItemListAction):
    """
    An action handler for crafting items
    """

    _command = "craft"

    def to_string(self):
        return "craft"

    def xml_template(self) -> str:
        return str("<SimpleCraftCommands/>")

    def __init__(self, items: list, _other=Optional[str], _default=Optional[str]):
        """
        Initializes the space of the handler to be one for each item in the list plus one for the
        default no-craft action (command 0)

        Items are minecraft resource ID's
        """
        kwargs = {}
        if _other is not None:
            kwargs["_other"] = _other
        if _default is not None:
            kwargs["_default"] = _default
        super().__init__(self._command, items, **kwargs)

    def from_universal(self, obs):
        if (
            "diff" in obs
            and "crafted" in obs["diff"]
            and len(obs["diff"]["crafted"]) > 0
        ):
            try:
                x = self._univ_items.index(obs["diff"]["crafted"][0]["item"])
                return obs["diff"]["crafted"][0]["item"].split("minecraft:")[-1]
            except ValueError:
                return self._default
                # return self._items.index('other')
        else:
            return self._default


class CraftWithTableAction(CraftAction):
    """
    An action handler for crafting items when agent is in view of a crafting table
    """

    _command = "craftNearby"

    def to_string(self):
        return "craft_with_table"

    def xml_template(self) -> str:
        return str("<NearbyCraftCommands/>")



import logging
import threading


class QueueLogger(logging.StreamHandler):
    def __init__(self, queue):
        self._queue = queue
        super().__init__(None)

    def flush(self):
        pass

    def emit(self, record):
        self._queue.append((self.level, record))


def launch_queue_logger_thread(output_producer, should_end):
    def queue_logger_thread(out_prod, should_end):
        while not should_end():
            try:
                line, running = out_prod.get_output()
                if not running:
                    break
                if line:
                    level = line[0]
                    record = line[1]
                    name = line[2]
                    lgr = logging.getLogger(name)
                    lgr.log(level, record)
            except Exception as e:
                print(e)
                break

    thread = threading.Thread(
        target=queue_logger_thread, args=(output_producer, should_end)
    )
    thread.setDaemon(True)
    thread.start()



import socket
import logging
import functools

import Pyro4


retry_count = 20
logger = logging.getLogger(__name__)


def retry(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        retry_exc = None
        for i in range(retry_count):
            try:
                return func(*args, **kwargs)
            except Pyro4.errors.PyroError as e:
                logger.error(
                    "An error occurred contacting the instance manager. Is it started!?"
                )
                raise e
            except (socket.timeout, socket.error, RuntimeError) as e:
                if retry_exc is None:
                    retry_exc = e
                if i < retry_count - 1:
                    logger.debug(f"Pause before retry on " + str(e))
                    raise e
        raise retry_exc

    return wrapper



from __future__ import annotations

from typing import Union, Dict, Optional, Tuple, List

from .base import ExtraSpawnMetaTaskBase
from ...sim.inventory import InventoryItem
from .extra_spawn import SpawnItem2Condition
from .utils import (
    always_satisfy_condition,
    use_all_item_check,
    use_any_item_reward,
    simple_inventory_based_reward,
)


ALL_TECH_LEVELS = [
    "wooden",
    "leather",
    "chainmail",
    "golden",
    "stone",
    "iron",
    "diamond",
    "archery",
    "explosives",
    "redstone",
]
ARMORS = ["boots", "chestplate", "helmet", "leggings"]
TOOLS = ["axe", "hoe", "pickaxe", "shovel", "sword"]
TECH_ITEMS = {
    "wooden": [f"wooden_{tool}" for tool in TOOLS],
    "leather": [f"leather_{armor}" for armor in ARMORS],
    "chainmail": [f"chainmail_{armor}" for armor in ARMORS],
    "golden": [f"golden_{item}" for item in ARMORS + TOOLS],
    "stone": [f"stone_{tool}" for tool in TOOLS],
    "iron": [f"iron_{item}" for item in ARMORS + TOOLS],
    "diamond": [f"diamond_{item}" for item in ARMORS + TOOLS],
    "archery": ["bow"],
    "explosives": ["tnt"],
    "redstone": [
        "redstone_block",
        "clock",
        "compass",
        "dispenser",
        "dropper",
        "observer",
        "piston",
        "redstone_lamp",
        "redstone_torch",
        "repeater",
        "detector_rail",
        "comparator",
        "activator_rail",
    ],
}
TECH_ITEMS_FLATTENED = [item for level in TECH_ITEMS.values() for item in level]


class TechTreeMeta(ExtraSpawnMetaTaskBase):
    """
    Class for tech tree tasks.
    Args:
        break_speed_multiplier: Controls the speed of breaking blocks. A value larger than 1.0 accelerates the breaking.
                Default: ``1.0``.

        event_level_control: If ``True``, the agent is able to perform high-level controls including place and equip.
                If ``False``, then is keyboard-mouse level control.
                Default: ``True``.

        fast_reset: If ``True``, reset using MC native command `/kill`, instead of waiting for re-generating new worlds.
            Default: ``True``.

            .. warning::
                Side effects:

                1. Changes to the world will not be reset. E.g., if the agent chops lots of trees then calling
                fast reset will not restore those trees.

                2. If you specify agent starting health and food, these specs will only be respected at the first reset
                (i.e., generating a new world) but will not be respected in the following resets (i.e., reset using MC cmds).
                So be careful to use this wrapper if your usages require specific initial health/food.

                3. Statistics/achievements will not be reset. This wrapper will maintain a property ``info_prev_reset``.
                If your tasks use stat/achievements to evaluation, please retrieve this property and compute differences.

        image_size: The size of image observations.

        initial_mobs: The types of mobs that are spawned initially.
                Default: ``None``.

        initial_mob_spawn_range_high: The upperbound on each horizontal axis from the center of the area to spawn initially.
                Default: ``None``.

        initial_mob_spawn_range_low: The lowerbound on each horizontal axis from the center of the area to spawn initially.
                Default: ``None``.

        initial_weather: If not ``None``, specifies the initial weather.
                Can be one of ``"clear"``, ``"normal"``, ``"rain"``, ``"thunder"``.
                Default: ``None``.

        lidar_rays: Defines the directions and maximum distances of the lidar rays if ``use_lidar`` is ``True``.
                If supplied, should be a list of tuple(pitch, yaw, distance).
                Pitch and yaw are in radians and relative to agent looking vector.
                Default: ``None``.

        obtain_items_reward_weights: The reward values of obtaining necessary items for unlocking the target tech.
                Default: ``1.0``.

        seed: The seed for an instance's internal generator.
                Default: ``None``.

        sim_name: Name of a simulation instance.
                Default: ``"TechTreeMeta"``.

        spawn_range_high: The upperbound on each horizontal axis from the center of the area to spawn
                Default: ``None``.

        spawn_range_low: The lowerbound on each horizontal axis from the center of the area to spawn
                Default: ``None``.

        spawn_rate: The probability of spawning in each step.
                Default: ``None``.

        start_food: If not ``None``, specifies initial food condition of the agent.
                Default: ``None``.

        start_health: If not ``None``, specifies initial health condition of the agent.
                Default: ``None``.

        start_position: If not ``None``, specifies the agent's initial location and orientation in the minecraft world.
                Default: ``None``.

        tech: Target tech to be unlocked.

        use_lidar: If ``True``, includes lidar in observations.
                Default: ``False``.

        use_items_reward_weights: The reward value of using the items, which is the last step of unlocking a tech.
                Default: ``10.0``.

        use_voxel: If ``True``, includes voxel in observations.
                Default: ``False``.

        voxel_size: Defines the voxel's range in each axis if ``use_voxel`` is ``True``.
                If supplied, should be a dict with keys ``xmin``, ``xmax``, ``ymin``, ``ymax``, ``zmin``, ``zmax``.
                Each value specifies the voxel size relative to the agent.
                Default: ``None``.

        world_seed: Seed for deterministic world generation
                if ``generate_world_type`` is ``"default"`` or ``"specified_biome"``.
                See `here <https://minecraft.fandom.com/wiki/Seed_(level_generation)>`_ for more details.
                Default: ``None``.
    """

    _prompt_template = "Unlock {tech} technology."

    def __init__(
        self,
        *,
        # ------ target technology ------
        tech: str | None = None,
        tech_item: str | None = None,
        obtain_items_reward_weights: Union[
            int, float, Dict[str, Union[int, float]]
        ] = 1.0,
        use_items_reward_weights: Union[
            int, float, Dict[str, Union[int, float]]
        ] = 10.0,
        # ------ initial & extra spawn control ------
        initial_mobs: Optional[Union[str, List[str]]] = None,
        initial_mob_spawn_range_low: Optional[Tuple[int, int, int]] = None,
        initial_mob_spawn_range_high: Optional[Tuple[int, int, int]] = None,
        spawn_rate: Optional[Dict[str, Union[float, int]]] = None,
        spawn_range_low: Optional[Tuple[int, int, int]] = None,
        spawn_range_high: Optional[Tuple[int, int, int]] = None,
        # ------ initial conditions ------
        initial_inventory: Optional[List[InventoryItem]] = None,
        start_position: Optional[Dict[str, Union[float, int]]] = None,
        start_health: Optional[float] = None,
        start_food: Optional[int] = None,
        initial_weather: Optional[str] = None,
        # ------ global conditions ------
        break_speed_multiplier: float = 1.0,
        # ------ sim seed & world seed ------
        seed: Optional[int] = None,
        world_seed: Optional[str] = None,
        # ------ reset mode ------
        fast_reset: bool = True,
        # ------ obs ------
        image_size: Union[int, Tuple[int, int]],
        use_voxel: bool = False,
        voxel_size: Optional[Dict[str, int]] = None,
        use_lidar: bool = False,
        lidar_rays: Optional[List[Tuple[float, float, float]]] = None,
        # ------ event-level action or keyboard-mouse level action ------
        event_level_control: bool = True,
        # ------ misc ------
        sim_name: str = "TechTreeMeta",
    ):
        # should provide `tech` or `tech_item`, but not both
        assert (tech is not None or tech_item is not None) and (
            tech is None or tech_item is None
        ), "should provide `tech` or `tech_item`, but not both"

        if tech is not None:
            assert (
                tech in ALL_TECH_LEVELS
            ), f"Invalid target tech: {tech}, expect one from {ALL_TECH_LEVELS}"
            self._target_tech = tech
            tech_items = TECH_ITEMS[tech]
        else:
            assert tech_item in TECH_ITEMS_FLATTENED
            self._target_tech = tech_item.replace("_", " ")
            tech_items = [tech_item]
        if isinstance(obtain_items_reward_weights, int) or isinstance(
            obtain_items_reward_weights, float
        ):
            obtain_items_reward_weights = {
                item: obtain_items_reward_weights for item in tech_items
            }
        elif isinstance(obtain_items_reward_weights, dict):
            assert set(obtain_items_reward_weights.keys()) == set(
                tech_items
            ), f"{set(obtain_items_reward_weights.keys())} must match {set(tech_items)}"
        if isinstance(use_items_reward_weights, int) or isinstance(
            use_items_reward_weights, float
        ):
            use_items_reward_weights = {
                item: use_items_reward_weights for item in tech_items
            }
        elif isinstance(use_items_reward_weights, dict):
            assert set(use_items_reward_weights.keys()) == set(
                tech_items
            ), f"{set(use_items_reward_weights.keys())} must match {set(tech_items)}"

        spawn_condition = None
        if spawn_rate is not None:
            assert isinstance(spawn_rate, dict), (
                f"in TechTree you must explicitly provide spawn items and rates by passing a dict, "
                f"but got {type(spawn_rate)}"
            )
            spawn_condition = {
                k: SpawnItem2Condition.get(k, always_satisfy_condition)
                for k in spawn_rate.keys()
            }

        # use any target items only *once* can succeed
        success_criteria = [use_all_item_check({k: 1 for k in tech_items})]
        reward_fns = [
            # rewards for obtaining target items
            simple_inventory_based_reward(name=k, weight=v)
            for k, v in obtain_items_reward_weights.items()
        ] + [
            # rewards for using target items
            use_any_item_reward(use_items_reward_weights)
        ]

        super().__init__(
            initial_mobs=initial_mobs,
            initial_mob_spawn_range_low=initial_mob_spawn_range_low,
            initial_mob_spawn_range_high=initial_mob_spawn_range_high,
            extra_spawn_rate=spawn_rate,
            extra_spawn_condition=spawn_condition,
            extra_spawn_range_low=spawn_range_low,
            extra_spawn_range_high=spawn_range_high,
            fast_reset=fast_reset,
            success_criteria=success_criteria,
            reward_fns=reward_fns,
            seed=seed,
            sim_name=sim_name,
            image_size=image_size,
            use_voxel=use_voxel,
            voxel_size=voxel_size,
            use_lidar=use_lidar,
            lidar_rays=lidar_rays,
            event_level_control=event_level_control,
            initial_inventory=initial_inventory,
            break_speed_multiplier=break_speed_multiplier,
            world_seed=world_seed,
            start_position=start_position,
            initial_weather=initial_weather,
            start_health=start_health,
            start_food=start_food,
        )

    @property
    def task_prompt(self) -> str:
        return super().get_prompt(tech=self._target_tech)



"""
A true flat inventory handler
"""
from typing import Dict, Any

import numpy as np

from minedojo.sim import spaces
from minedojo.sim.handlers.translation import TranslationHandler


class TrueFlatInventoryObservation(TranslationHandler):
    n_slots = 36
    # exclude equipment slots
    excluded_slots = [36, 37, 38, 39, 40]

    def to_string(self) -> str:
        return "inventory"

    def xml_template(self) -> str:
        return str("""<ObservationFromFullInventory flat="false"/>""")

    def __init__(self):
        shape = (self.n_slots,)
        space = spaces.Dict(
            {
                "name": spaces.Text(shape=shape),
                # max variant index is 120, i.e., item id 383:120 spawn egg for villager
                # see https://minecraft-ids.grahamedgecombe.com/
                "variant": spaces.MultiDiscrete([121 for _ in range(self.n_slots)]),
                # min stack size is 0 (no item), max stack size is 64
                "quantity": spaces.Box(
                    low=0, high=64, shape=(self.n_slots,), dtype=np.float32
                ),
                # though durability in MC seems to be integer, we still use box with float32 here
                "max_durability": spaces.Box(
                    low=-1, high=np.inf, shape=shape, dtype=np.float32
                ),
                "cur_durability": spaces.Box(
                    low=-1, high=np.inf, shape=shape, dtype=np.float32
                ),
            }
        )
        super().__init__(space=space)

    def from_hero(self, obs_dict: Dict[str, Any]):
        assert "inventory" in obs_dict, "Missing inventory key in malmo json"
        raw_inventory = [
            item for item in obs_dict["inventory"] if item["inventory"] == "inventory"
        ]
        assert len(raw_inventory) == self.n_slots + len(self.excluded_slots), "INTERNAL"
        return {
            key: np.array(
                [
                    self._name_preprocess(slot_dict[key])
                    if key == "name"
                    else slot_dict[key]
                    for i, slot_dict in enumerate(raw_inventory)
                    if i not in self.excluded_slots
                ],
                dtype=self.space[key].dtype,
            ).reshape(self.space[key].shape)
            for key in self.space
        }

    @staticmethod
    def _name_preprocess(raw_name: str):
        if "_" in raw_name:
            return raw_name.replace("_", " ")
        else:
            return raw_name


class EquipmentObservation(TrueFlatInventoryObservation):
    n_slots = 6
    excluded_slots = list(range(1, 36))

    def to_string(self) -> str:
        return "equipment"



"""
If there is a crafting table or furnace nearby
"""
from typing import Any

import numpy as np

from minedojo.sim import spaces
from minedojo.sim.handlers.translation import TranslationHandler


class NearbyToolsObservation(TranslationHandler):
    def to_string(self) -> str:
        return "nearby_tools"

    def xml_template(self) -> str:
        return ""

    def __init__(self):
        space = spaces.Dict(
            {
                "table": spaces.Box(low=0, high=1, dtype=bool, shape=()),
                "furnace": spaces.Box(low=0, high=1, dtype=bool, shape=()),
            }
        )
        super().__init__(space=space)

    def from_hero(self, x: dict[str, Any]):
        return {
            "table": np.array(x["nearby_crafting_table"], dtype=bool),
            "furnace": np.array(x["nearby_furnace"], dtype=bool),
        }



from __future__ import annotations

import json
import os

from .download import download as dl
from .download import get_fn


class YouTubeDataset:
    """
    Class for MineDojo YouTube Database API.
    We follow PyTorch Dataset format but without actually inheriting from PyTorch dataset to keep the framework general.

    Args:
        download: If set to ``True`` and there is no existing cache directory, the data will be downloaded automatically.

        download_dir: Directory path where the downloaded data will be saved.
                Default: ``~/.minedojo/``.

        full: If ``True``, the full version of the YouTube database will be downlaoded.
                If ``False``, only the tutorial version of the YouTube database will be downloaded.
                Default: ``True``.

    Examples:
        >>> from minedojo.data import YouTubeDataset
        >>> youtube_dataset = YouTubeDataset()
        >>> print(youtube_dataset[0].keys())
        dict_keys(['id', 'title', 'link', 'view_count', 'like_count', 'duration', 'fps'])
    """

    def __init__(
        self,
        *,
        download: bool = True,
        download_dir: None | str = None,
        full: bool = True,
    ):
        if download_dir is None:
            download_dir = os.path.join(os.path.expanduser("~"), ".minedojo")

        if download:
            self.root = dl("youtube", download_dir, full)
        else:
            self.root, _, url = get_fn("youtube", download_dir, full)
            assert os.path.exists(self.root), (
                f"YouTube data file {self.root} does not exist. "
                "Please set download=True or you can manually "
                f"download it from {url}."
            )

        with open(self.root, "r") as f:
            self.data = json.load(f)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.data[idx]
